///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Common/AssemblyHelper/MetadataIterators.cpp 7     5/11/06 10:18a Eaddy $
//
//    File name:	MetadataIterators.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Common/AssemblyHelper/MetadataIterators.cpp $
 * 
 * 7     5/11/06 10:18a Eaddy
 * o Fixed error msg
 * 
 * 6     5/02/06 11:47p Eaddy
 * o Fixed bug where we were looking up the token in the wrong metadata
 * import causing incorrect behavior and an infinite recursion in rare
 * cases
 * 
 * 5     4/22/06 12:35p Eaddy
 * o Adding support for using the Debugger API to query for the current
 * metadata and IL
 * 
 * 4     4/14/06 5:36p Eaddy
 * o Trying to fix emitting methods
 * 
 * 3     4/12/06 7:39p Eaddy
 * o Made slightly more flexible
 * 
 * 2     4/11/06 8:29p Eaddy
 * o Added new emit method that assumes the existing type def parent token
 * 
 * 1     3/16/06 1:53a Eaddy
 * 
 * 1     3/16/06 1:49a Eaddy
 * 
 * 16    1/20/06 9:39p Eaddy
 * o Migrated to VS 2005.  Replaced many function calls with their safe
 * version.
 * 
 * 15    12/14/05 2:16a Eaddy
 * o Even if we don't support emitting new types, methods, and fields, 
 * we still need to fix them up.  And we can't assume that they can
 * be iterated in lock step since they may have been reordered.
 * Actually, this only really happens when the build description file
 * doesn't match the executable because the exe was built using
 * VS.
 * 
 * 14    12/09/05 9:33p Eaddy
 * o Added support for introducing fields but of course this doesn't work
 * either because the CLR requires DefineField to be called in
 * ModuleLoadFinished
 * 
 * 13    12/09/05 9:03p Eaddy
 * o Minor changes
 * 
 * 12    12/09/05 4:34p Eaddy
 * o methodDef_iterator now uses CMethodInfo to store method props
 * o Now all iterators use the same ctor signature to make the
 * MetadataDiff template method happy and also to ensure vital
 * params are available to subtypes
 * 
 * 11    12/08/05 10:34p Eaddy
 * o Now really fixed problem with changing a method's local variables
 * 
 * 10    12/08/05 8:29p Eaddy
 * o Now compares signatures correctly
 * o Now supports changing a method's local variables
 * 
 * 9     12/07/05 11:44p Eaddy
 * o Fixed bug so now assembly refs are emitted correctly
 * 
 * 8     12/07/05 11:36p Eaddy
 * o Added assemblyRef iter
 * 
 * 7     12/07/05 7:54p Eaddy
 * o Added iterators for typerefs and member refs
 * o Added emit()
 * o Now each iterator provides a way to obtain, compare, and print its
 * metadata value
 * 
 * 6     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.2  2004/10/10 06:07:45  me133
 * o Added header comment
 */

#include "StdAfx.h"

#include "MetadataIterators.h"	// Me

#include <corerror.h>			// CLDB_S_TRUNCATION

////////////////////////////////
// ASSEMBLY REF IMPLEMENTATION
////////////////////////////////

assemblyRef_iterator::assemblyRef_iterator(IMetaDataImport* piImport)
	: _Mybase(piImport)
	, m_spiAssemblyImport()
	, m_assemblyProps()
{ 
	HRESULT hr = piImport->QueryInterface(	IID_IMetaDataAssemblyImport, 
		(void**) &m_spiAssemblyImport.p );
	VERIFY_HR_COND_V(hr, m_spiAssemblyImport, _T("QI for IMetaDataAssemblyImport failed."), ILXLOG_ERROR);

	getNextBatch();
}

assemblyRef_iterator::assemblyRef_iterator(IMetaDataImport* piImport, mdToken tokParent)
	: _Mybase(piImport, tokParent)
	, m_spiAssemblyImport()
	, m_assemblyProps()
{ 
	HRESULT hr = piImport->QueryInterface(	IID_IMetaDataAssemblyImport, 
		(void**) &m_spiAssemblyImport.p );
	VERIFY_HR_COND_V(hr, m_spiAssemblyImport, _T("QI for IMetaDataAssemblyImport failed."), ILXLOG_ERROR);

	getNextBatch();
}

void assemblyRef_iterator::initCurrent()
{
	HRESULT hr = S_OK;

	ULONG nActualNameLen = 0;

	hr = m_spiAssemblyImport->GetAssemblyRefProps(
		getToken(),							// [IN]  The AssemblyRef for which to get the properties.
		&m_assemblyProps.pbPubKeyOrTok,		// [OUT] Pointer to the public key or token.
		&m_assemblyProps.nPubKeyLen,		// [OUT] Count of bytes in the public key or token.
		m_assemblyProps.name,				// [OUT] Buffer to fill with name.
		sizeof(m_assemblyProps.name)/sizeof(*m_assemblyProps.name)-1,	// [IN] Size of buffer in wide chars.
		&nActualNameLen,					// [OUT] Actual # of wide chars in name.
		&m_assemblyProps.assemblyMetadata,	// [OUT] Assembly MetaData.
		&m_assemblyProps.pbHashBlob,		// [OUT] Hash blob.
		&m_assemblyProps.nHashLen,			// [OUT] Count of bytes in the hash blob.
		&m_assemblyProps.dwFlags);			// [OUT] Flags.

	CHECK_HR(hr, _T("Failed to retrieve assembly ref properties."), ILXLOG_ERROR);
}

const WCHAR* assemblyRef_iterator::getName() const
{
	return m_assemblyProps.toString();
}

const SAssemblyProps& assemblyRef_iterator::getValue() const
{
	return m_assemblyProps;
}

HRESULT assemblyRef_iterator::emit(IMetaDataEmit* piMetaDataEmit, 
								   mdToken /*tokEnclosingScope*/, 
								   mdToken* ptokEmitted)
{
	HRESULT hr = S_OK;

	CComPtr<IMetaDataAssemblyEmit> spiAssemblyEmit;

	hr = piMetaDataEmit->QueryInterface(	IID_IMetaDataAssemblyEmit, 
											(void**) &spiAssemblyEmit.p );
	VERIFY_HR_COND(hr, m_spiAssemblyImport, _T("QI for IMetaDataAssemblyEmit failed."), ILXLOG_ERROR);

	hr = spiAssemblyEmit->DefineAssemblyRef(m_assemblyProps.pbPubKeyOrTok,
											m_assemblyProps.nPubKeyLen,
											m_assemblyProps.name,
											&m_assemblyProps.assemblyMetadata,
											m_assemblyProps.pbHashBlob,
											m_assemblyProps.nHashLen,
											m_assemblyProps.dwFlags,
											ptokEmitted);
	return hr;
}

void assemblyRef_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiAssemblyImport->EnumAssemblyRefs(	&m_hEnum,
												m_ary, 
												_countof(m_ary), 
												&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumAssemblyRefs() failed."), ILXLOG_ERROR);
}

/////////////////////////////////////
// TYPE DEF ITERATOR IMPLEMENTATION
/////////////////////////////////////

#pragma warning (disable:4351) // elements of array will be default initialized
typeDef_iterator::typeDef_iterator(IMetaDataImport* piImport)
	: _Mybase(piImport)
	, m_szTypeName()
	, m_dwFlags(0)
	, m_tokBaseType(mdTokenNil)
{ 
	getNextBatch();
}

typeDef_iterator::typeDef_iterator(IMetaDataImport* piImport, mdToken tokParent)
	: _Mybase(piImport, tokParent)
	, m_szTypeName()
	, m_dwFlags(0)
	, m_tokBaseType(mdTokenNil)
{ 
	getNextBatch();
}
#pragma warning (default:4351)

void typeDef_iterator::initCurrent()
{
	getProps(	m_spiImport, 
				this->operator*(), 
				m_szTypeName, 
				(sizeof(m_szTypeName)/sizeof(*m_szTypeName))-1, 
				&m_dwFlags, 
				&m_tokBaseType,
                0);
}

HRESULT typeDef_iterator::getProps(	IMetaDataImport* piImport,
									mdTypeDef tokTypeDef, 
									WCHAR* szTypeName, 
									ULONG nMaxLen, 
									DWORD* pdwFlags, 
									mdToken* ptokBaseType,
                                    IMetaDataImport** ppiNewImport)
{
	HRESULT hr = S_OK;

    _ASSERTE(piImport && piImport);

	szTypeName[0] = L'\0';

	ULONG cchTypeName = 0;

	if (TypeFromToken(tokTypeDef) == mdtTypeRef)
	{
		CComPtr<IMetaDataImport> spiImportTypeRef;

        mdTypeDef tokTypeDefCopy = tokTypeDef;

		hr = piImport->ResolveTypeRef(tokTypeDefCopy, IID_IMetaDataImport, 
			(IUnknown**)(&spiImportTypeRef.p), &tokTypeDef); 
 		VERIFY_HR_COND1(hr, spiImportTypeRef.p, ILXLOG_ERROR, 
            _T("ResolveTypeRef(0x%08X) failed."), tokTypeDefCopy);
		
		hr = spiImportTypeRef->GetTypeDefProps(	tokTypeDef, 
												szTypeName, 
												nMaxLen, 
												&cchTypeName, 
												pdwFlags,
												ptokBaseType);
        if (ppiNewImport)
        {
            _ASSERTE(!*ppiNewImport); // Forget to Release()?
            *ppiNewImport = spiImportTypeRef;
            (*ppiNewImport)->AddRef();
        }
    }
	else 
	{
		// This is expensive
		hr = piImport->GetTypeDefProps(		tokTypeDef, 
											szTypeName, 
											nMaxLen, 
											&cchTypeName, 
											pdwFlags, 
											ptokBaseType);
        if (ppiNewImport)
        {
            _ASSERTE(!*ppiNewImport); // Forget to Release()?
            *ppiNewImport = piImport;
            (*ppiNewImport)->AddRef();
        }
    }

	VERIFY_HR1(hr, ILXLOG_ERROR, _T("GetTypeDefProps(0x%08X) failed."), tokTypeDef);
	if (cchTypeName > nMaxLen)
	{
		return ILXLOG3(E_UNEXPECTED, ILXLOG_ERROR, _T("Type name for 0x%08X too big (%d), max is %d."), 
			tokTypeDef, cchTypeName, nMaxLen);
	}

	// For some reason we can't retrieve the type name for System.Object
	// JScript.NET doesn't have types
	//_ASSERTE(*szTypeName || (cchTypeName == 0 && hr == S_FALSE));

	return hr;
}

bool typeDef_iterator::isDelegate() const
{
	HRESULT hr = S_OK;

	// Traverse up the inheritance hierarchy to see if we are derived from
	// System.MulticastDelegate

	mdToken tokBaseType = m_tokBaseType;

    int count = 0;

    CComPtr<IMetaDataImport> spiImport;
    spiImport = m_spiImport;

	while (	hr == S_OK && 
			TypeFromToken(tokBaseType) != mdTokenNil && 
			!IsNilToken(tokBaseType))
	{
		const int nMaxLen = 2048; // 2048 from alink.h
		WCHAR szBaseTypeName[nMaxLen+1]; 

        CComPtr<IMetaDataImport> spiNewImport;
		hr = getProps(spiImport, tokBaseType, szBaseTypeName, nMaxLen, 0, &tokBaseType, &spiNewImport);

        spiImport = spiNewImport;

		// An optimization would be to store this result. However, isDelegate() is only
		// ever called once for the current iteration
		if (hr == S_OK && wcscmp(szBaseTypeName, L"System.MulticastDelegate") == 0)
			return true;

        ++count;
        if (count > 1000)
        {
            ILXLOGMSG1(ILXLOG_ERROR, _T("Exceeded infinite loop threshold (%d)."), 
                count);
            _ASSERTE(false);
            return false;
        }
	}
	
	return false;
}

bool typeDef_iterator::isInterface() const
{
	return IsTdInterface(m_dwFlags);
}

const SStrProp typeDef_iterator::getValue() const
{
	return SStrProp(m_szTypeName);
}

const WCHAR* typeDef_iterator::getName() const
{
	return m_szTypeName;
}

void typeDef_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumTypeDefs( &m_hEnum,
									m_ary, 
									_countof(m_ary), 
									&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumTypeDefs() failed."), ILXLOG_ERROR);
}

HRESULT typeDef_iterator::emit(IMetaDataEmit* piMetaDataEmit, mdToken tokEnclosingScope, mdToken* ptokEmitted)
{
	return E_NOTIMPL;
}

/////////////////////////////////////
// TYPE REF ITERATOR IMPLEMENTATION
/////////////////////////////////////

#pragma warning (disable:4351) // elements of array will be default initialized
typeRef_iterator::typeRef_iterator(IMetaDataImport* piImport)
	: _Mybase(piImport)
	, m_szTypeName()
	, m_tokResolutionScope(mdTokenNil)
{ 
	getNextBatch();
}

typeRef_iterator::typeRef_iterator(IMetaDataImport* piImport, mdToken tokParent)
	: _Mybase(piImport, tokParent)
	, m_szTypeName()
	, m_tokResolutionScope(mdTokenNil)
{ 
	getNextBatch();
}
#pragma warning (default:4351)

void typeRef_iterator::initCurrent()
{
	getProps(m_spiImport, this->operator*(), m_szTypeName, 
		(sizeof(m_szTypeName)/sizeof(*m_szTypeName))-1, &m_tokResolutionScope);

	if (m_tokParent != mdTokenNil && m_tokResolutionScope != m_tokParent)
	{
		this->operator++();
	}
}

HRESULT typeRef_iterator::getProps(IMetaDataImport* piImport, 
								   mdToken tokType, WCHAR* szTypeName, ULONG nMaxLen,
								   mdToken* ptokResolutionScope)
{
	_ASSERTE(piImport && ptokResolutionScope);

	HRESULT hr = S_OK;

	szTypeName[0] = L'\0';

	ULONG cchTypeName = 0;

	mdToken tokResolutionScope = mdTokenNil;

	hr = piImport->GetTypeRefProps(	tokType, 
		ptokResolutionScope,
		szTypeName, 
		nMaxLen, 
		&cchTypeName);

	VERIFY_HR1(hr, ILXLOG_ERROR, _T("GetTypeRefProps(0x%08X) failed."), tokType);
	if (cchTypeName > nMaxLen)
	{
		return ILXLOG3(E_UNEXPECTED, ILXLOG_ERROR, _T("Type name for 0x%08X too big (%d), max is %d."), tokType, cchTypeName, nMaxLen);
	}

	// For some reason we can't retrieve the type name for System.Object
	// JScript.NET doesn't have types
	//_ASSERTE(*szTypeName || (cchTypeName == 0 && hr == S_FALSE));

	return hr;
}

const SStrProp typeRef_iterator::getValue() const
{
	return SStrProp(m_szTypeName);
}

const WCHAR* typeRef_iterator::getName() const
{
	return m_szTypeName;
}

void typeRef_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumTypeRefs( &m_hEnum,
		m_ary, 
		_countof(m_ary), 
		&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumTypeRefs() failed."), ILXLOG_ERROR);
}

HRESULT typeRef_iterator::emit(IMetaDataEmit* piMetaDataEmit, 
							   mdToken tokEnclosingScope, 
							   mdToken* ptokEmitted)
{
	HRESULT hr = S_OK;

	hr = piMetaDataEmit->DefineTypeRefByName(tokEnclosingScope,
		m_szTypeName, ptokEmitted);
	VERIFY_HR1(hr, ILXLOG_ERROR, _T("Failed to emit type ref '%ls'."), m_szTypeName);
	return hr;
}

///////////////////////////////////////
// METHOD DEF ITERATOR IMPLEMENTATION
///////////////////////////////////////

methodDef_iterator::methodDef_iterator(IMetaDataImport* piImport, mdToken tokParent)
	: _Mybase(piImport, tokParent)
	, m_methodInfo()
    , m_pilFetcher(0)
{
	getNextBatch();
}

methodDef_iterator::methodDef_iterator(typeDef_iterator& itTypeDef, 
									   const CILFetcher& ilFetcher)
	: _Mybase(itTypeDef)
	, m_methodInfo()
    , m_pilFetcher(&ilFetcher)
{
	getNextBatch();
}

methodDef_iterator::methodDef_iterator(typeDef_iterator& itTypeDef) 
	: _Mybase(itTypeDef)
	, m_methodInfo()
    , m_pilFetcher(0)
{
	getNextBatch();
}

const CMethodInfo& methodDef_iterator::getValue() const
{
	return m_methodInfo;
}

CMethodInfo& methodDef_iterator::getValue()
{
	return m_methodInfo;
}

const WCHAR* methodDef_iterator::getParentName() const
{
	HRESULT hr = S_OK;

	_ASSERTE(m_tokParent != mdTokenNil);

	static WCHAR szParentName[2048*2];

	szParentName[0] = L'\0';

	DWORD dwFlagsIgnore = 0;
	mdToken tokBaseTypeIgnore = mdTokenNil;

	hr = typeDef_iterator::getProps(	m_spiImport,
										m_tokParent, 
										szParentName, 
										(sizeof(szParentName) / sizeof(szParentName[0])-1),
										&dwFlagsIgnore,
										&tokBaseTypeIgnore,
                                        0);

	CHECK_HR1(hr, ILXLOG_ERROR, _T("Failed to get properties for type def 0x%08X."), 
		m_tokParent);

	return szParentName;
}

void methodDef_iterator::initCurrent()
{
	m_methodInfo.init(m_spiImport, this->operator*(), 0, m_pilFetcher);

	_ASSERTE(m_tokParent == mdTokenNil || m_tokParent == getValue().m_typeDef);
	m_tokParent = m_methodInfo.m_typeDef;
}

void methodDef_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumMethods( &m_hEnum,
									m_tokParent,
									m_ary, 
									_countof(m_ary), 
									&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumMethods() failed."), ILXLOG_ERROR);
}

HRESULT methodDef_iterator::emit(IMetaDataEmit* piMetaDataEmit, 
								 mdToken tokEnclosingScope, 
								 mdToken* ptokEmitted)
{
	return m_methodInfo.emit(piMetaDataEmit, tokEnclosingScope, ptokEmitted);
}

HRESULT methodDef_iterator::emit(IMetaDataEmit* piMetaDataEmit, mdToken* ptokEmitted)
{
	return m_methodInfo.emit(piMetaDataEmit, ptokEmitted);
}

///////////////////////////////////////
// MEMBER REF ITERATOR IMPLEMENTATION
///////////////////////////////////////

memberRef_iterator::memberRef_iterator(IMetaDataImport* piImport, mdToken tokParent)
	: _Mybase(piImport, tokParent)
	, m_props()
{
	// No need ref count since lifetime is controlled by parent
	// (memberRef_iterator)
	m_props.piImport = m_spiImport; 

	getNextBatch();
}

memberRef_iterator::memberRef_iterator(IMetaDataImport* piImport, mdToken tok, bool)
	: _Mybase(piImport, mdTokenNil)
	, m_props()
{
	// No need ref count since lifetime is controlled by parent
	// (memberRef_iterator)
	m_props.piImport = m_spiImport; 

	m_ary[0] = tok;
	m_ulTotal = 1;
	m_ulCount = 1;
	initCurrent();
}

memberRef_iterator::memberRef_iterator(typeRef_iterator& itTypeRef) 
	: _Mybase(itTypeRef)
	, m_props()
{
	// No need ref count since lifetime is controlled by parent
	// (memberRef_iterator)
	m_props.piImport = m_spiImport; 

	getNextBatch();
}

void memberRef_iterator::initCurrent()
{
	getProps(this->operator*(), m_props.name, (sizeof(m_props.name)/sizeof(*m_props.name))-1, 
		m_props.fullName, (sizeof(m_props.fullName)/sizeof(*m_props.fullName))-1, 
		m_tokParent, &m_props.sig, &m_props.siglen);
}

HRESULT memberRef_iterator::getProps(	mdMemberRef tokMemberRef, 
										WCHAR* szMemberName, 
										ULONG nMaxLen, 
										WCHAR* szFullName, 
										ULONG nMaxFullNameLen,
										const mdTypeRef& tokEnclosingTypeRef, 
										PCCOR_SIGNATURE* ppSignature, 
										ULONG* pnSignatureLength) const
{
	HRESULT hr = S_OK;

	_ASSERTE(tokMemberRef != mdTokenNil);
	_ASSERTE(szMemberName);
	_ASSERTE(nMaxLen > 0);
	_ASSERTE(ppSignature);
	_ASSERTE(pnSignatureLength);

	szMemberName[0] = L'\0';

	ULONG cchMemberName = 0;

	mdTypeRef tokEnclosingTypeRefToCheck = mdTypeRefNil;

	hr = m_spiImport->GetMemberRefProps(	tokMemberRef, 
											&tokEnclosingTypeRefToCheck, 
											szMemberName, 
											nMaxLen, 
											&cchMemberName,
											ppSignature, 
											pnSignatureLength) ;

	_ASSERTE(tokEnclosingTypeRef == tokEnclosingTypeRefToCheck || 
		tokEnclosingTypeRef == mdTokenNil);

	wcscpy_s(szFullName, nMaxFullNameLen, getParentName());
	wcscat_s(szFullName, nMaxFullNameLen, L".");
	wcscat_s(szFullName, nMaxFullNameLen, getName());

	return hr;
}

const SMemberProps& memberRef_iterator::getValue() const
{
	return m_props;
}

const WCHAR* memberRef_iterator::getName() const
{
	return m_props.name;
}

const WCHAR* memberRef_iterator::getParentName() const
{
	HRESULT hr = S_OK;

	_ASSERTE(m_tokParent != mdTokenNil);

	static WCHAR szParentName[2048*2];
	
	szParentName[0] = L'\0';

	mdToken tokResolutionScope = mdTokenNil;

	hr = typeRef_iterator::getProps(m_spiImport,
									m_tokParent, 
									szParentName, 
									(sizeof(szParentName) / sizeof(szParentName[0])-1),
									&tokResolutionScope);
	
	CHECK_HR1(hr, ILXLOG_ERROR, _T("Failed to get properties for type ref 0x%08X."), 
		m_tokParent);

	return szParentName;
}

const WCHAR* memberRef_iterator::getFullName() const
{
	return m_props.fullName;
}

void memberRef_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumMemberRefs(	&m_hEnum,
										m_tokParent,
										m_ary, 
										_countof(m_ary), 
										&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumMethods() failed."), ILXLOG_ERROR);
}

HRESULT memberRef_iterator::emit(	IMetaDataEmit* piMetaDataEmit,
									mdToken tokEnclosingScope,
									mdToken* ptokEmitted)
{
	_ASSERTE(tokEnclosingScope != mdTokenNil);

	return piMetaDataEmit->DefineMemberRef(	tokEnclosingScope, 
											m_props.name,
											m_props.sig,
											m_props.siglen,
											ptokEmitted);
}

///////////////////////////////////////
// FIELD DEF ITERATOR IMPLEMENTATION
///////////////////////////////////////

fieldDef_iterator::fieldDef_iterator(IMetaDataImport* piImport, mdToken tokParent)
	: _Mybase(piImport, tokParent)
	, m_props()
{
	// No need ref count since lifetime is controlled by parent
	// (fieldDef_iterator)
	m_props.piImport = m_spiImport; 

	getNextBatch();
}

fieldDef_iterator::fieldDef_iterator(typeRef_iterator& itTypeRef) 
	: _Mybase(itTypeRef)
	, m_props()
{
	// No need ref count since lifetime is controlled by parent
	// (fieldDef_iterator)
	m_props.piImport = m_spiImport; 

	getNextBatch();
}

void fieldDef_iterator::initCurrent()
{
	HRESULT hr = S_OK;

	mdTypeDef tokEnclosingClass = mdTypeDefNil;

	ULONG cchMemberName = 0;

	hr = m_spiImport->GetFieldProps(getToken(),
									&tokEnclosingClass,
									m_props.name,
									(sizeof(m_props.name)/sizeof(*m_props.name))-1,
									&cchMemberName,
									&m_props.attrs,
									&m_props.sig,
									&m_props.siglen,
									&m_props.cplus,
									&m_props.val,
									&m_props.valsize);
	VERIFY_HR1_V(hr, ILXLOG_ERROR, _T("GetFieldProps(0x%08X) failed."), getToken());

	_ASSERTE(tokEnclosingClass == m_tokParent);

	wcscpy_s(m_props.fullName, getParentName());
	wcscat_s(m_props.fullName, L".");
	wcscat_s(m_props.fullName, getName());
}

const SFieldProps& fieldDef_iterator::getValue() const
{
	return m_props;
}

const WCHAR* fieldDef_iterator::getName() const
{
	return m_props.name;
}

const WCHAR* fieldDef_iterator::getParentName() const
{
	HRESULT hr = S_OK;

	_ASSERTE(m_tokParent != mdTokenNil);

	static WCHAR szParentName[2048*2];

	szParentName[0] = L'\0';

	DWORD dwFlagsIgnore = 0;
	mdToken tokBaseTypeIgnore = mdTokenNil;

	hr = typeDef_iterator::getProps(	m_spiImport,
										m_tokParent, 
										szParentName, 
										(sizeof(szParentName) / sizeof(szParentName[0])-1),
										&dwFlagsIgnore,
										&tokBaseTypeIgnore,
                                        0);

	CHECK_HR1(hr, ILXLOG_ERROR, _T("Failed to get properties for type def 0x%08X."), 
		m_tokParent);

	return szParentName;
}

const WCHAR* fieldDef_iterator::getFullName() const
{
	return m_props.fullName;
}

void fieldDef_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumFields(	&m_hEnum,
									m_tokParent,
									m_ary, 
									_countof(m_ary), 
									&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumMethods() failed."), ILXLOG_ERROR);
}

HRESULT fieldDef_iterator::emit(	IMetaDataEmit* piMetaDataEmit,
									mdToken tokEnclosingScope,
									mdToken* ptokEmitted)
{
	return E_NOTIMPL;
/*
	_ASSERTE(tokEnclosingScope != mdTokenNil);

	// Apparently, this only works when called from ModuleLoadFinished
	return piMetaDataEmit->DefineField(	tokEnclosingScope, 
										m_props.name,
										m_props.attrs,
										m_props.sig,
										m_props.siglen,
										m_props.cplus,
										m_props.val,
										m_props.valsize,
										ptokEmitted);
*/
}

//////////////////////////////////////
// PARAM DEF ITERATOR IMPLEMENTATION
//////////////////////////////////////

paramDef_iterator::paramDef_iterator(IMetaDataImport* piImport, mdToken tokParent)
	: _Mybase(piImport, tokParent)
{
	getNextBatch();
}

paramDef_iterator::paramDef_iterator(methodDef_iterator& itMethodDef) 
	: _Mybase(itMethodDef)
{
	getNextBatch();
}

void paramDef_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumParams(	&m_hEnum,
									m_tokParent,
									m_ary, 
									_countof(m_ary), 
									&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumParams() failed."), ILXLOG_ERROR);
}

////////////////////////////////////////
// USER STRING ITERATOR IMPLEMENTATION
////////////////////////////////////////

userString_iterator::userString_iterator(IMetaDataImport* piImport, mdToken tokParent)
	: _Mybase(piImport, tokParent)
	, m_pszTemp(0)
	, m_ulSize(0)
{
	reserve(nINITIAL_SIZE);
	getNextBatch();
}

userString_iterator::~userString_iterator()
{
	free(m_pszTemp);
}

// The string returned will be invalid if you increment or
// destroy the iterator
const SStrWLen& userString_iterator::getValue() const
{
	static SStrWLen strTemp;

	HRESULT hr = S_OK;

	if (end())
	{
		_ASSERTE(!"End of iteration");
		strTemp.str.clear();
		strTemp.len = -1;
		return strTemp;
	}

	ULONG chUserString = 0;           // Size of user string.

    while (true) 
	{ 
		// Try to get the string into the existing buffer.
		// Subtract 1 because we need the extra space for the
		// null terminator
        hr = m_spiImport->GetUserString(	this->operator*(), 
											m_pszTemp,
											m_ulSize-1, 
											&chUserString);
        if (hr == CLDB_S_TRUNCATION)
        {   
			// Buffer wasn't big enough, try to enlarge it.
			reserve(chUserString);
        }
		else
		{
			// Buffer was big enough
			break;
		}
    }

	if (m_pszTemp && chUserString >= 0 && chUserString < m_ulSize)
	{
		m_pszTemp[chUserString] = L'\0';
	}

	strTemp.str = m_pszTemp;
	strTemp.len = chUserString;

	CHECK_HR(hr, _T("GetUserString() failed."), ILXLOG_ERROR);

	return strTemp;
}

void userString_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumUserStrings(	&m_hEnum,
										m_ary, 
										_countof(m_ary), 
										&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumUserStrings() failed."), ILXLOG_ERROR);
}

void userString_iterator::reserve(ULONG nCapacity) const
{
	// Add one for the null terminator which metadata doesn't set
	nCapacity = max(nCapacity, m_ulSize * nINC_FACTOR) + 1;

	// Buffer wasn't big enough, try to enlarge it
	free(m_pszTemp);
	m_pszTemp = (WCHAR*) malloc(nCapacity * sizeof(WCHAR));
	m_ulSize = nCapacity;
}

HRESULT userString_iterator::emit(IMetaDataEmit* piMetaDataEmit, 
								  mdToken /*tokEnclosingScope*/, 
								  mdToken* ptokEmitted)
{
	const SStrWLen& str = getValue();
	return piMetaDataEmit->DefineUserString(	str.toString(),
												str.len,
												ptokEmitted);
}

//////////////////////////////////////
// SIGNATURE ITERATOR IMPLEMENTATION
//////////////////////////////////////

signature_iterator::signature_iterator(IMetaDataImport* piImport, mdToken tokParent)
	: _Mybase(piImport, tokParent)
{ 
	getNextBatch();
}

void signature_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumSignatures(	&m_hEnum,
										m_ary, 
										_countof(m_ary), 
										&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumSignatures() failed."), ILXLOG_ERROR);
}

void signature_iterator::initCurrent()
{
	HRESULT hr = m_spiImport->GetSigFromToken(	this->operator*(),
												&m_props.sig,
												&m_props.len);
	VERIFY_HR_V(hr, _T("GetSigFromToken() failed."), ILXLOG_ERROR);

	// No ref count since props is owned by parent (signature_iterator)
	m_props.piImport = m_spiImport.p;

	m_props.init();
}

const SSigProps& signature_iterator::getValue()
{
	return m_props;
}

HRESULT signature_iterator::emit(IMetaDataEmit* piMetaDataEmit, 
								 mdToken /*tokEnclosingScope*/, 
								 mdToken* ptokEmitted)
{
	_ASSERTE(piMetaDataEmit);

	HRESULT hr = S_OK;

	hr = piMetaDataEmit->GetTokenFromSig(m_props.sig, m_props.len, ptokEmitted);
	VERIFY_HR(hr, _T("GetTokenFromSig() failed."), ILXLOG_ERROR);

	return hr;
}
