//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/MorphPlugin/TypeMorphPhase.cs 12    6/02/06 2:10a Eaddy $
//
// Description:
//
//    Master phase for controlling assembly morphing.
//
// Remarks:
//
//    TypeMorphPhase is responsible for morphing types. It inserts
//    itself immediately after the PE Reader phase.  It also creates
//    the WeavingPhase which is responsible for weaving aspects at
//    the FuncUnit level.
//
// Specifying extensions:
//
//    TypeMorphPhase looks in the reference assemblies for extension
//    custom attributes.  The following extension attributes are
//    supported:
//
//       AddAttribute - Add a field or method
//       ReplaceAttribute - Replace a method
//       AdviceAttribute - Aspect-Oriented Programming-style advice
//       MorphExecAttribute - Arbitrary code that will be executed at
//          morph-time via System.Reflection
//
// Example:
//
//    Add a bool field 'b' to class HelloNS.Hello:
//
//       using Phx.Morph.Attributes;
//       public class MyExt
//       {
//          [Add("HelloNS.Hello")]
//          public bool b;
//       }
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/MorphPlugin/TypeMorphPhase.cs $
 * 
 * 12    6/02/06 2:10a Eaddy
 * o Got rid of redundant verbosity flags
 * 
 * 11    5/24/06 9:48p Eaddy
 * o Now prints diagnostic messages in color
 * o Verbosity is now static
 * 
 * 10    5/05/06 6:18p Eaddy
 * o Did some renaming to improve clarity
 * o Can now list current advice and enable/disable (although
 * enable/disable doesn't actually work yet)
 * 
 * 9     5/03/06 5:06p Eaddy
 * o Fixed bug where attributes that were defined in the module being
 * woven
 * were not accessible via System.Reflection
 * 
 * 8     5/03/06 12:24a Eaddy
 * o Made names a little more obvious
 * o Fixed null ref exception
 * o No longer checks for aspects in a module that's already been checked
 * o Fixed bug where verbosity level changed behavior
 * 
 * 7     5/02/06 5:06p Eaddy
 * o Optimizing dynamic weaving
 * 
 * 6     5/02/06 4:51p Eaddy
 * o Made aspects/advice static so we can avoid loading it multiple times
 * in the dynamic weaving scenario
 * 
 * 5     5/02/06 12:12p Eaddy
 * o Did not need to pass moduleUnitRefs
 * 
 * 4     4/19/06 1:39a Eaddy
 * o Made verbosity level more fine-grained
 * 
 * 3     3/15/06 12:37p Eaddy
 * o Added verbosity flag and ShowSummary() method
 * 
 * 2     2/24/06 8:19p Eaddy
 * o Now uses Phx.Aop.dll
 */

// ISSUE-TODO-t-marcea-2005/08/01 - The reason why TypeMorphPhase
// creates WeavingPhase is so they can share assembly references
// and the array of extension attributes.  They should be decoupled.

using System;

using System.Collections.Generic;

using Dbg = System.Diagnostics.Debug;

using Phx.Morph.Editors;

namespace Phx.Morph
{
   /// <summary>
   /// Master phase for controlling assembly morphing
   /// </summary>
   /// <remarks>
   /// TypeMorphPhase is responsible for morphing types. It inserts
   /// itself immediately after the PE Reader phase.  It also creates
   /// the WeavingPhase which is responsible for weaving aspects at
   /// the FuncUnit level.
   /// </remarks>

   public class TypeMorphPhase : Phx.Phases.Phase
   {
      /// <summary>
      /// Extension attributes collected from the reference assemblies
      /// and/or command-line
      /// </summary>

      static List<Phx.Morph.Attributes.IExtensionAttribute> aryExtensions =
         new List<Phx.Morph.Attributes.IExtensionAttribute>();

      /// <summary>
      /// Extension and reference assemblies
      /// </summary>
      /// <remarks>
      /// The assemblies will be used for locating extension
      /// attributes and for resolving types used by the
      /// extensions
      /// <br>Specified using /r on the command-line</br>
      /// </remarks>

      static List<string> aspectCandidates = new List<string>();
      static List<string> candidatesAlreadyLoaded = new List<string>();

      static bool somethingChanged = false;

      public static bool SomethingChanged
      {
         get { return somethingChanged || WeavingPhase.SomethingChanged; }
      }

      /// <summary>
      /// Weaves aspect code at the FuncUnit level
      /// </summary>

      WeavingPhase weavingPhase = null;

      /// <summary>
      /// Count of errors
      /// </summary>

      int errorCount = 0;

      static TypeMorphPhase()
      {
         // Dummy call to force CLR to load Phx.Aop.Morph so we can use
         // its attribute types
         new Phx.Aop.Attributes.AddAttribute();

         // Force the initialization of the attribute handler registry
         Phx.Morph.Attributes.AttributeHandlerRegistry.InitStatic();
      }

      public static string Name
      {
         get { return "Type Morph Phase"; }
      }

      /// <summary>
      /// Initializes and inserts the TypeMorphPhase,
      /// along with a WeavingPhase, into the phase list
      /// </summary>
      /// <param name="config">PhaseList configuration to insert into</param>
      /// <param name="references">Extension and reference assemblies
      /// (specified using the /r command-line option)</param>
      /// <param name="debugCtrl">Debug control to use</param>
      /// <returns>true on success</returns>

      public bool
      Setup
      (
         Phx.Phases.PhaseConfig config,
         string[] references,
         Phx.Ctrls.CompCtrl debugCtrl
      )
      {
         this.Init(config, Name);

         string phaseName = "PE Reader";

         Phx.Phases.Phase phaseFind = config.PhaseList.FindByName(phaseName);
         if (phaseFind == null)
         {
            Morpher.WriteError("'{0}' not found in phase list: '{1}'",
               phaseName, config.NameString);
            return false;
         }

         phaseFind.InsertAfter(this);

         this.weavingPhase = new WeavingPhase();
         this.weavingPhase.Setup(config, debugCtrl);

         //config.PhaseList.Dump();
         //
         //Morpher Phases (no ctrl) [PhaseList]
         //   PE Reader (ctrl pereader)
         //   Type Morph Phase (no ctrl)
         //   Ir Unit Func Phases (no ctrl) [PhaseList]
         //      RaiseIR (ctrl RaiseIR)
         //      Weaving Phase (no ctrl)
         //      Stackification (ctrl MSILStackify)
         //      Encode (no ctrl)
         //   PE Writer (ctrl pewriter)

         // Initialize our assembly references

         if (references != null)
         {
            TypeMorphPhase.aspectCandidates.AddRange(references);
         }

         // Initialize the debug control

         PhaseCtrl = debugCtrl;

         return true;
      }

      /// <summary>
      /// Morphs the assembly
      /// </summary>
      /// <param name="unit">PEModuleUnit to morph</param>

      override protected void
      Execute
      (
         Phx.Unit unit
      )
      {
         somethingChanged = false;

         Phx.PEModuleUnit moduleUnit = unit as Phx.PEModuleUnit;
         Dbg.Assert(moduleUnit != null);
         if (moduleUnit == null)
         {
            ++this.errorCount;
            return;
         }

         // I don't think we need to load the moduleUnit using Reflection
         if (!candidatesAlreadyLoaded.Contains(moduleUnit.NameString))
         {
            ReflectionHelpers.LoadAssembly(moduleUnit.NameString);
            candidatesAlreadyLoaded.Add(moduleUnit.NameString);
         }

         // Load and parse the aspect assemblies

         foreach(string reference in TypeMorphPhase.aspectCandidates)
         {
            if (candidatesAlreadyLoaded.Contains(reference))
                continue;

            TypeMorphPhase.LoadAspectCandidate(reference);
            candidatesAlreadyLoaded.Add(reference);
         }

         // Clear the list of references so we don't load them again
         TypeMorphPhase.aspectCandidates.Clear();

         this.weavingPhase.SetExtensions(TypeMorphPhase.aryExtensions);

         // See if the extension instructions are embedded in attributes

         if (TypeMorphPhase.aryExtensions.Count == 0)
         {
            Morpher.WriteWarning("No type extensions found.");
            return;
         }

         ApplyExtensions(moduleUnit, TypeMorphPhase.aryExtensions);
      }

      /// <summary>
      /// Adds the reference assembly to be loaded during Execute()
      /// </summary>
      /// <remarks>
      /// We load the assemblies into Phoenix so we can locate
      /// extension attributes and resolve types specified in
      /// the extensions.  Because in some cases we use .NET
      /// Reflection on these types (and for MorphExec) we also
      /// load the assemblies into the runtime (using
      /// System.Reflection.Assembly.Load).
      /// </remarks>
      /// <param name="references">Reference assembly</param>

      public static void AddAspectCandidate
      (
         string aspectCandidatePath
      )
      {
         aspectCandidates.Add(aspectCandidatePath);
      }

      /// <summary>
      /// Loads the reference assemblie into Phoenix and CLR
      /// </summary>
      /// <remarks>
      /// We load the assemblies into Phoenix so we can locate
      /// extension attributes and resolve types specified in
      /// the extensions.  Because in some cases we use .NET
      /// Reflection on these types (and for MorphExec) we also
      /// load the assemblies into the runtime (using
      /// System.Reflection.Assembly.Load).
      /// </remarks>
      /// <param name="references">Reference assembly</param>

      static void LoadAspectCandidate
      (
         string aspectCandidatePath
      )
      {
         // Load the reference module

         Phx.PEModuleUnit moduleUnitRef = 
            ReflectionHelpers.LoadModuleUnit(aspectCandidatePath);
         if (moduleUnitRef == null)
         {
            Morpher.WriteError("Failed to open reference/aspect assembly '{0}'.",
               aspectCandidatePath);
            return;
         }

         // This load the assembly so we can use reflection on it later

         ReflectionHelpers.LoadAssembly(aspectCandidatePath);

         Morpher.WriteOutput("Loaded reference/aspect assembly '{0}'.",
            aspectCandidatePath);

         // Used for instantiating a custom attribute using its
         // metadata blob

         Phx.Aop.Attributes.AttributeHelper attributeHelper =
            new Phx.Aop.Attributes.AttributeHelper();

         attributeHelper.LocateExtensions(moduleUnitRef, ref TypeMorphPhase.aryExtensions);
      }

      /// <summary>
      /// Morphs the module according to the module extension attributes
      /// </summary>
      /// <remarks>
      /// Non-module extension attributes (not derived from
      /// IModuleMorphAttribute) are ignored
      /// </remarks>
      /// <param name="moduleUnit">Module to morph</param>
      /// <param name="aryExtensions">Extension attributes to apply</param>

      void
      ApplyExtensions
      (
         Phx.PEModuleUnit moduleUnit,
         List<Phx.Morph.Attributes.IExtensionAttribute> aryExtensions
      )
      {
         // A ModuleEditor is used to simplify modifying the module

         ModuleEditor moduleEditor = new ModuleEditor(moduleUnit);

         foreach (Phx.Morph.Attributes.IExtensionAttribute extensionAttribute in
            aryExtensions)
         {
            if (!extensionAttribute.IsInited())
            {
               // Command-line extensions (e.g., /field:System.Boolean,b)
               // must be initialized now

               if (!extensionAttribute.Init(null, moduleUnit.TypeTable))
               {
                  ++this.errorCount;
                  continue;
               }
            }

            Dbg.Assert(extensionAttribute.IsInited());

            Phx.Morph.Attributes.IModuleMorphAttribute moduleMorphAttribute =
               extensionAttribute as Phx.Morph.Attributes.IModuleMorphAttribute;
            if (moduleMorphAttribute == null)
            {
               // Not a module extension attribute

               continue;
            }

            // The "business logic" for morphing is specific for each
            // attribute and so is handled by the attribute itself

            bool success = moduleMorphAttribute.Morph(moduleEditor);
            if (!success)
            {
               ++this.errorCount;
            }
            else
            {
               somethingChanged = true;
            }
         }
      }

      /// <summary>
      /// Count of errors encountered during Execute()
      /// </summary>

      public int
      ErrorCount
      {
         get { return this.errorCount + this.weavingPhase.ErrorCount; }
      }

      public void
      ShowSummary()
      {
         if (weavingPhase != null)
            weavingPhase.ShowSummary();
      }
   }
}
