//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/MorphPlugin/WeavingPhase.cs 14    6/02/06 2:10a Eaddy $
//
// Description:
//
//    Phase that performs function-level (FuncUnit) aspect weaving.
//
// Remarks:
//
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/MorphPlugin/WeavingPhase.cs $
 * 
 * 14    6/02/06 2:10a Eaddy
 * o Got rid of redundant verbosity flags
 * 
 * 13    5/24/06 9:48p Eaddy
 * o Now prints diagnostic messages in color
 * o Verbosity is now static
 * 
 * 12    5/24/06 2:03p Eaddy
 * o Changed method name
 * 
 * 11    5/24/06 1:35a Eaddy
 * o Switched names for MethodWeaver and MethodEditor
 * so now they make sense
 * o Fixed bug where AppendFuncArg didn't work when
 * weaving static methods
 * 
 * 10    5/05/06 6:18p Eaddy
 * o Did some renaming to improve clarity
 * o Can now list current advice and enable/disable (although
 * enable/disable doesn't actually work yet)
 * 
 * 9     5/03/06 12:26a Eaddy
 * o Now figures out if something was changed during weaving
 * o Advice is now static so we only load it and apply it once
 * 
 * 8     5/02/06 12:12p Eaddy
 * o Did not need to pass moduleUnitRefs
 * 
 * 7     4/19/06 1:39a Eaddy
 * o Made verbosity level more fine-grained
 * 
 * 6     3/15/06 2:35p Eaddy
 * o Now uses weaver factory so we can choose from breakpoint
 * or binary weaving
 * 
 * 5     3/15/06 12:53p Eaddy
 * o Moved message to ShowSummary
 * 
 * 4     3/15/06 12:38p Eaddy
 * o Added verbosity flag and ShowSummary() method
 * o Optimized advice initialization
 * 
 * 3     3/01/06 2:21a Eaddy
 * o Now uses new Phx.Aop library
 * 
 * 2     2/24/06 8:19p Eaddy
 * o Now uses Phx.Aop.dll
 */

using System;

using System.Collections;
using System.Collections.Generic;

using Dbg = System.Diagnostics.Debug;

using Phx.Morph.Editors;

namespace Phx.Morph
{
   /// <summary>
   /// Phase that performs function-level (FuncUnit) aspect weaving
   /// </summary>

   public class WeavingPhase : Phx.Phases.Phase
   {
      /// <summary>
      /// Advice info extracted from the advice extension attributes
      /// </summary>

      static List<Phx.Aop.PhxAdvice> advices = null;

      /// <summary>
      /// Count of errors
      /// </summary>

      int errorCount = 0;

      static bool somethingChanged;

      public static bool SomethingChanged
      {
         get { return somethingChanged; }
      }

      public static IEnumerable Advices
      {
         get { return advices; }
      }

      /// <summary>
      /// Creates a new WeavingPhase
      /// </summary>
      /// <remarks>
      /// Call in this order: 1) New, 2) Setup, 
      /// 3) SetReferencesAndExtensions, 4) Execute
      /// </remarks>
      /// <returns>A new WeavingPhase</returns>

      public static WeavingPhase
      New
      (
      )
      {
         return new WeavingPhase();
      }

      /// <summary>
      /// Initializes and inserts the WeavingPhase,
      /// along with some dependent phases, into the phase list
      /// </summary>
      /// <param name="config">PhaseList configuration to insert into</param>
      /// <param name="debugCtrl">Debug control to use</param>
      /// <remarks>
      /// Call in this order: 1) New, 2) Setup, 
      /// 3) SetReferencesAndExtensions, 4) Execute
      /// </remarks>
      /// <returns>true on success</returns>

      public bool
      Setup
      (
         Phx.Phases.PhaseConfig config,
         Phx.Ctrls.CompCtrl debugCtrl
      )
      {
         // So we don't call Setup() twice

         Dbg.Assert(this.PhaseCtrl == null);

         this.Init(config, "Weaving Phase");

         // Find the FuncUnit phase list

         string phaseName = "Ir Unit Func Phases";

         Phx.Phases.Phase phaseFind = config.PhaseList.FindByName(phaseName);
         if (phaseFind == null)
         {
            Morpher.WriteError("'{0}' not found in phase list: '{1}'",
               phaseName, config.NameString);
            return false;
         }

         Phx.PE.UnitListPhaseList funcUnitPhaseList =
            phaseFind as Phx.PE.UnitListPhaseList;

         Dbg.Assert(phaseFind != null);
         if (phaseFind == null)
         {
            return false;
         }

         // Insert ourselves after the RaiseIR phase

         phaseName = "RaiseIR";
         phaseFind = funcUnitPhaseList.FindByName(phaseName);
         if (phaseFind == null)
         {
            Morpher.WriteError("'{0}' not found in phase list: '{1}'",
               phaseName, config.NameString);
            return false;
         }

         phaseFind.InsertAfter(this);

         //this.InsertAfter(Phx.StackAllocPhase.New(config));

         //this.InsertAfter(Phx.Targets.Runtimes.FrameGenerationPhase.New(config));

         //this.InsertAfter(Phx.Targets.Archs.MSIL.StackificationPhase.New(config));

         //config.PhaseList.Dump();
         //
         //Morpher Phases (no ctrl) [PhaseList]
         //   PE Reader (ctrl pereader)
         //   Type Morph Phase (no ctrl)
         //   Ir Unit Func Phases (no ctrl) [PhaseList]
         //      RaiseIR (ctrl RaiseIR)
         //      Weaving Phase (no ctrl)
         //      Stackification (ctrl MSILStackify)
         //      Frame Generation (ctrl Frame)
         //      Stack Allocation (ctrl StackAlloc)
         //      Encode (no ctrl)
         //   PE Writer (ctrl pewriter)

         // Initialize the debug control

         PhaseCtrl = debugCtrl;

         // Clear the applied flag so we can determine if something
         // new was woven

         if (WeavingPhase.advices != null)
         {
            foreach (Phx.Aop.PhxAdvice advice in WeavingPhase.advices)
            {
               advice.AppliedToNewJoinpoints = false;
            }
         }

         return true;
      }

      /// <summary>
      /// Sets the extension attributes used to control the weaving
      /// process and resolve types
      /// </summary>
      /// <remarks>
      /// Call in this order: 1) New, 2) Setup, 3) SetExtensions,
      ///   4) Execute
      /// </remarks>
      /// <param name="aryExtensions">Extension attributes</param>

      public void
      SetExtensions
      (
         List<Phx.Morph.Attributes.IExtensionAttribute> aryExtensions
      )
      {
         Dbg.Assert(aryExtensions != null);

         // Gather the pointcuts from all the AdviceAttributes

         foreach (Phx.Morph.Attributes.IExtensionAttribute extensionAttribute in
            aryExtensions)
         {
            Phx.Morph.Attributes.AdviceAttributeHandler adviceAttribute =
               extensionAttribute as Phx.Morph.Attributes.AdviceAttributeHandler;

            if (adviceAttribute == null || adviceAttribute.Advice == null)
            {
               // Not an advice attribute
               continue;
            }

            bool found = false;

            if (advices == null)
            {
               advices = new List<Phx.Aop.PhxAdvice>();
            }
            else
            {
               foreach (Phx.Aop.PhxAdvice advice in WeavingPhase.advices)
               {
                  if (advice == adviceAttribute.Advice)
                     found = true;
               }
            }

            if (!found)
               WeavingPhase.advices.Add(adviceAttribute.Advice);
         }
      }

      //--------------------------------------------------------------------------
      //
      // Description:
      //
      //--------------------------------------------------------------------------

      /// <summary>
      /// Weaves the function
      /// </summary>
      /// <param name="unit">FuncUnit to weave</param>
      /// <remarks>
      /// Call in this order: 1) New, 2) Setup, 
      /// 3) SetReferencesAndExtensions, 4) Execute
      /// </remarks>

      override protected void
      Execute
      (
         Phx.Unit unit
      )
      {
         somethingChanged = false;

         Dbg.Assert(unit != null);
         if (unit == null)
         {
            return;
         }

         // Skip data units

         if (unit.IsDataUnit)
         {
            return;
         }

         Phx.FuncUnit funcUnit = unit as Phx.FuncUnit;
         Dbg.Assert(funcUnit != null);

         // ISSUE-TODO-AndyA-2005/01/21 Need a better way to address
         // target dependent plugin behavior.

         // Only handle MSIL functions.

         if (!funcUnit.Arch.NameString.Equals("MSIL"))
         {
            return;
         }

         ApplyExtensions(funcUnit, WeavingPhase.advices);
      }

      /// <summary>
      /// Morphs the function according to the advice extension attributes
      /// </summary>
      /// <remarks>
      /// Non-advice extension attributes (not derived from
      /// IAdviceAttribute) are ignored
      /// </remarks>
      /// <param name="funcUnit">Function to morph</param>
      /// <param name="aryExtensions">Advice extension attributes to apply</param>

      void
      ApplyExtensions
      (
         Phx.FuncUnit funcUnit,
         List<Phx.Aop.PhxAdvice> advices
      )
      {
         Dbg.Assert(funcUnit != null && funcUnit.FuncSym != null);
         if (funcUnit == null || funcUnit.FuncSym == null)
         {
            ++this.errorCount;
            return;
         }

         // Make sure we have advice
         if (WeavingPhase.advices == null || WeavingPhase.advices.Count == 0)
         {
            return;
         }

         // Skip non-MSIL methods (e.g. UnmanagedEntryPoint)

         if (funcUnit.FuncSym.EnclosingAggrType == null)
         {
            return;
         }

         // A MethodEditor does the actual weaving

         MethodEditor methodWeaveCandidate = 
            MethodEditorFactory.CreateEditor(funcUnit.FuncSym);

         // ISSUE-TODO-t-marcea-2005/08/02- From a design standpoint,
         // it makes more sense to pass in a list of advice to the
         // Weave() method.  Advice=pointcut+methodToCall.

         this.errorCount += methodWeaveCandidate.Weave(advices);
      }

      /// <summary>
      /// Count of errors encountered during Execute()
      /// </summary>

      public int
      ErrorCount
      {
         get { return this.errorCount; }
      }

      public void
      ShowSummary()
      {
         somethingChanged = false;

         if (WeavingPhase.advices != null && WeavingPhase.advices.Count > 0)
         {
            foreach (Phx.Aop.PhxAdvice advice in WeavingPhase.advices)
            {
               if (!advice.Enabled)
               {
                  Morpher.WriteOutput("Advice '{0}' was disabled.",
                     advice);
               }
               else if (advice.AppliedToNewJoinpoints)
               {
                  somethingChanged = true;
               }
               else
               {
                  Morpher.WriteWarning("Advice '{0}' was never applied.",
                     advice);
               }
            }
         }
         else
         {
            Morpher.WriteWarning("No advice found.");
         }
      }
   }
}
