//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Aop/Attributes/ExtendsAttribute.cs 3     2/24/06 8:17p Eaddy $
//
// Description:
//
//    Marks that the class contains extensions and specifies
//    which class its extending.  Any ifaces or base class
//    implemented by the attached class will be added
//    automatically.
//
// Example:
//
//       [Extends("Hello")]
//       class MyExt
//       {
//          [Add]
//          bool b;
//       }
//
// Remarks:
//
//    All extension classes must have this attribute attached
//    to the class.  When, later in the class definition, we
//    encounter an Add, Replace, etc. attribute we will find
//    the parent Extends attribute which will tell us which
//    class we are extending.
//
//    Handles adding:
//
//       Base classes
//       Base interfaces
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Aop/Attributes/ExtendsAttribute.cs $
 * 
 * 3     2/24/06 8:17p Eaddy
 * o Initial checkin
 */

using System;

namespace Phx.Aop.Attributes
{
   /// <summary>
   /// Marks that the class contains extensions and specifies
   /// which class its extending.  Any ifaces or base class
   /// implemented by the attached class will be added
   /// automatically.
   /// </summary>
   /// <remarks>
   /// All extension classes must have this attribute attached
   /// to the class.  When, later in the class definition, we
   /// encounter an Add, Replace, etc. attribute we will find
   /// the parent Extends attribute which will tell us which
   /// class we are extending.
   /// </remarks>

   [AttributeUsage(AttributeTargets.Class, AllowMultiple = true)]
   public class ExtendsAttribute : Attribute
   {
      // ISSUE-BUG-t-marcea-2005/08/12-I'd like to support
      // specifying the System.Type instead of the FQN for
      // a class since it would allow us to catch more errors
      // at compile-time.  I had some problems, however. One
      // is that if I have one ctor that takes a string and one
      // that takes a System.Type, the compiler seems to pick
      // ctor(string) and implicitly calls System.Type.ToString().
      // Even after I removed ctor(string), when I try to
      // instantiate the attribute I get an exception that
      // Node_Visitor.dll couldn't be found.

      /// <summary>
      /// Fully qualified name of the class we will be morphing.
      /// </summary>
      /// <example>
      /// In this example, the extends attribute is attached to Bar
      /// and extends Foo:
      /// <code>
      ///   [Extends("Foo")]
      ///   class Bar { ... }
      /// </code>
      /// </example>

      string typeToExtendString = null;

      /// <summary>
      /// Keeps track of all the instantiated ExtendsAttribute
      /// objects
      /// </summary>
      /// <remarks>
      /// This allows us to lookup an extends attribute
      /// later by specifying the FQN of the attached class
      /// </remarks>
      /// <seealso cref="GetExtendsAttribute"/>

      public ExtendsAttribute
      (
         string typeToExtendString
      )
      {
         this.typeToExtendString = typeToExtendString;
      }

      /// <summary>
      /// Fully qualified name of the type this attribute
      /// extends
      /// </summary>

      public string TypeToExtendString
      {
         get { return this.typeToExtendString; }
         set { this.typeToExtendString = value; }
      }
   }
}
