//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Aop/Attributes/StatementAttribute.cs 12    5/24/06 9:47p Eaddy $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Aop/Attributes/StatementAttribute.cs $
 * 
 * 12    5/24/06 9:47p Eaddy
 * o Better diagnostics when the ctor param list for the inner
 * attribute doesn't match any ctor
 * o Now tries to resolve assemblies that can't be loaded
 * 
 * 11    5/11/06 11:26p Eaddy
 * o Breaks out of the loop early
 * o Checks for null
 * 
 * 10    5/11/06 7:48p Rr2318
 * Does not require parity between arguments and fields
 * 
 * 8     5/10/06 5:18p Rr2318
 * Added safety check to ensure StatementAttribute.Hydrate()'s chosen
 * constructor's arguments match supplied arguments
 * 
 * 7     5/05/06 12:40p Eaddy
 * o Now uses a type spec so that we load assemblies appropriately
 * 
 * 6     5/04/06 9:51p Eaddy
 * o Now returns the ConstructorInfo used
 * 
 * 5     5/04/06 3:04p Eaddy
 * o Changed return type of Attributes property so that we aren't
 * locked in to how we implement it
 * 
 * 4     5/04/06 2:20p Eaddy
 * o Removed unused field
 * 
 * 3     5/04/06 2:19p Eaddy
 * o Removed empty ctor
 * 
 * 2     5/04/06 2:19p Eaddy
 * o Modified ToString()
 * 
 * 1     5/03/06 3:18p Rr2318
 * Initial Check-in
 */

using System;
using System.IO;

using Phx.Aop.Pointcuts;

namespace Phx.Aop.Attributes
{
   /// <summary>
   /// Handles weaving advice into a method
   /// </summary>

   [AttributeUsage(AttributeTargets.Method | AttributeTargets.Constructor, AllowMultiple = true)]
   public class StatementAttribute : Attribute
   {
      uint originalSourceLine;
      string attrTypeSpec;
      System.Attribute inner = null;
      object[] innerAttrCtorArgs;
      System.Reflection.ConstructorInfo ctorInfo = null;

      public StatementAttribute(StatementAttribute that)
      {
         this.originalSourceLine = that.originalSourceLine;
         this.attrTypeSpec = that.attrTypeSpec;
         this.inner = that.inner;
         this.innerAttrCtorArgs = that.innerAttrCtorArgs;
         this.ctorInfo = that.ctorInfo;
      }

      public StatementAttribute
      (
         uint originalSourceLine,
         string innerAttrTypeSpec,
         params object[] innerAttrCtorArgs
      )
      {
         this.originalSourceLine = originalSourceLine;
         this.attrTypeSpec = innerAttrTypeSpec;
         this.innerAttrCtorArgs = innerAttrCtorArgs;
      }

      public Attribute Hydrate()
      {
         System.Type attrType = MyGetType(attrTypeSpec);
         if (attrType == null)
            return null;

         Type[] argTypes;

         if (innerAttrCtorArgs.Length > 0)
         {
            argTypes = new Type[innerAttrCtorArgs.Length];
            for (int i = 0; i < innerAttrCtorArgs.Length; i++)
            {
               argTypes[i] = innerAttrCtorArgs[i].GetType();
            }
         }
         else
         {
            argTypes = System.Type.EmptyTypes;
         }

         ctorInfo = attrType.GetConstructor(argTypes);

         if (ctorInfo == null)
         {
            System.Reflection.ConstructorInfo[] ci = attrType.GetConstructors();
            for (int i = 0; ci != null && i < ci.Length; ++i)
            {
               System.Reflection.ParameterInfo[] pi = ci[i].GetParameters();
               if (pi == null || pi.Length != argTypes.Length) 
                  continue;

               bool match = true;
               for (int j = 0; j < pi.Length; ++j)
               {
                  if(argTypes[j].ToString() != pi[j].ParameterType.FullName) 
                  {
                     match = false;
                     break;
                  }
               }

               if (match) 
               { 
                  ctorInfo = ci[i];
                  break;
               } 
            }
         }

         if (ctorInfo == null)
         {
            System.Text.StringBuilder ctorsGot = new System.Text.StringBuilder();
            System.Text.StringBuilder ctorsExpected = new System.Text.StringBuilder();

            ctorsGot.Append('(');

            for (int i = 0; i < argTypes.Length; ++i)
            {
               if (i > 0)
                  ctorsGot.Append(", " + argTypes[i]);
               else
                  ctorsGot.Append(argTypes[i]);
            }

            ctorsGot.Append(')');

            System.Reflection.ConstructorInfo[] ci = attrType.GetConstructors();
            for (int i = 0; ci != null && i < ci.Length; ++i)
            {
               if (i > 0)
                  ctorsExpected.Append(", ");

               if (i > 1 && i == ci.Length-1)
                  ctorsExpected.Append("or ");

               ctorsExpected.Append('(');

               System.Reflection.ParameterInfo[] pi = ci[i].GetParameters();
               if (pi != null)
               {
                  for (int j = 0; j < pi.Length; j++)
                  {
                     if (j > 0)
                        ctorsExpected.Append(", " + pi[j].ParameterType);
                     else
                        ctorsExpected.Append(pi[j].ParameterType);
                  }
               }

               ctorsExpected.Append(')');
            }

            throw new System.FormatException(string.Format(
               "Could not find a matching attribute constructor. Got {0}, but expected {1}.",
               ctorsGot,
               ctorsExpected));
         }

         object newobj = ctorInfo.Invoke(innerAttrCtorArgs);

         inner = newobj as Attribute;

         return inner;
      }
      
      public override string ToString()
      {
         System.Text.StringBuilder sb = new System.Text.StringBuilder();
         sb.Append( "Line: " + originalSourceLine + " Inner:");

         if (inner != null) 
         {
            sb.Append(" " + inner.ToString());
         }
         else
         {
            sb.Append(" null");
         }

         sb.Append(" (" + attrTypeSpec.ToString() + ")");

         return sb.ToString();
      }
      
      public System.Attribute InnerAttribute
      {
         get { return this.inner; }
      }

      public System.Reflection.ConstructorInfo CtorInfo
      {
         get 
         {
            System.Diagnostics.Debug.Assert(ctorInfo != null, "Must call Hydrate() first.");
            return ctorInfo;
         }
      }

      public uint Line
      {
         get { return this.originalSourceLine; }
         set { this.originalSourceLine = value; }
      }

      public string AttributeType
      {
         get 
         {
            int comma = this.attrTypeSpec.IndexOf(',');
            if (comma == -1)
            {
               return attrTypeSpec;
            }
            else
            {
               return attrTypeSpec.Substring(0, comma);
            }
         }
      }

      public string AttributeTypeSpec
      {
         get { return this.attrTypeSpec; }
         set { this.attrTypeSpec = value; }
      }

      public object[] Args
      {
         get { return this.innerAttrCtorArgs; }
         set { this.innerAttrCtorArgs = value; }
      }

      /// <summary>
      /// Obtains the Type for a fully qualified type name by looking
      /// in all the loaded assemblies
      /// </summary>
      /// <remarks>
      /// Type.GetType(string) is supposed to do this exact thing
      /// but it doesn't work
      /// </remarks>
      /// <param name="typeName">Fully qualified name of the type
      /// (System.String)</param>
      /// <returns>Requested Type or null</returns>

      static public Type
      MyGetType
      (
         string typeName
      )
      {
         // See if the typename includes an assembly name,
         // e.g., MyType,MyAssembly
         int comma = typeName.IndexOf(',');
         string assemblyName = null;
         if (comma > 0)
         {
            assemblyName = typeName.Substring(comma+1);
         }

         // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
         // for unmanaged implementation

         System.Type clrType = System.Type.GetType(typeName);
         if (clrType != null)
         {
            return clrType;
         }

         // Scour the other loaded assemblies for the type. This is
         // exactly what Type.GetType() is supposed to do (IMHO).

         bool assemblyFound = false;

         System.AppDomain clrDomain = System.AppDomain.CurrentDomain;
         System.Reflection.Assembly[] clrAssemblies = clrDomain.GetAssemblies();
         foreach (System.Reflection.Assembly clrAssembly in clrAssemblies)
         {
            if (assemblyName != null)
            {
               string curAssemblyName;
               comma = clrAssembly.FullName.IndexOf(',');
               if (comma > 0)
               {
                  curAssemblyName = clrAssembly.FullName.Substring(0, comma);
               }
               else
               {
                  curAssemblyName = clrAssembly.FullName;
               }

               // If the user supplied an assembly name then use it
               // to filter out which assemblies we search
               if (curAssemblyName != assemblyName)
                  continue;

               assemblyFound = true;
            }

            System.Reflection.Module[] clrModules = clrAssembly.GetModules();
            foreach (System.Reflection.Module clrModule in clrModules)
            {
               clrType = clrModule.GetType(typeName);
               if (clrType != null)
                  return clrType;
            }
         }

         if (assemblyName != null && !assemblyFound)
         {
            System.Reflection.Assembly clrAssembly = LoadAssembly(assemblyName);
            if (clrAssembly == null)
               return null; // Failed to load the assembly

            System.Reflection.Module[] clrModules = clrAssembly.GetModules();
            foreach (System.Reflection.Module clrModule in clrModules)
            {
               clrType = clrModule.GetType(typeName);
               if (clrType != null)
                  return clrType;
            }
         }

         // We did not find the CLR type because the module where the type
         // is defined has not been loaded via System.Reflection.  Try calling
         // ReflectionHelpers.LoadAssembly() first.
         return null;
      }

      /// <summary>
      /// Load an assembly into the CLR
      /// </summary>
      /// <param name="assemblyString">
      /// Can be an assembly short name (assembly name without
      /// an extension), relative path, or full path.
      /// </param>
      /// <returns>
      /// Loaded CLR assembly or null if the assembly couldn't be loaded
      /// </returns>

      public static System.Reflection.Assembly
      LoadAssembly
      (
         string assemblyString
      )
      {
         System.AppDomain currentDomain = System.Threading.Thread.GetDomain();
         currentDomain.AssemblyResolve += AssemblyResolveEventHandler;

         // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
         // for unmanaged implementation

         // Strip off the extension if present

         try
         {
            if (Path.IsPathRooted(assemblyString))
            {
               return System.Reflection.Assembly.LoadFrom(assemblyString);
            }
            else
            {
               switch (Path.GetExtension(assemblyString))
               {
                  // ISSUE-TODO-t-marcea-2005/07/23-Load() and LoadFile()
                  // don't work if an extension is specified (verify?).
                  // However, the cases below probably don't cover all
                  // possibilities.

                  case ".dll":
                  case ".exe":
                  case ".netmodule":
                     assemblyString = Path.Combine(
                        Path.GetDirectoryName(assemblyString),
                        Path.GetFileNameWithoutExtension(assemblyString));
                     break;
               }

               return System.Reflection.Assembly.Load(assemblyString);
            }
         }
         catch (Exception ex)
         {
            System.Console.WriteLine(
               "Phx.Morph: Warning: Assembly load failed for '{0}': {1}",
               assemblyString,
               ex.Message);
            return null;
         }
      }

      public static System.Reflection.Assembly
      AssemblyResolveEventHandler
      (
         object sender,
         ResolveEventArgs e
      )
      {
         string assemblyName = e.Name;

         int comma = assemblyName.IndexOf(',');
         if (comma > 0)
         {
            assemblyName = assemblyName.Substring(0, comma);
         }

         string[] tryExts = { "", ".dll", ".exe", ".netmodule" };
         foreach (string ext in tryExts)
         {
            try
            {
               return System.Reflection.Assembly.LoadFrom(assemblyName + ext);
            }
            catch (Exception)
            { }
         }

         return null;
      }
   }
}
