//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Aop/Pointcuts/Expressions/NestedPointcutExpression.cs 9     5/11/06 12:44p Eaddy $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Aop/Pointcuts/Expressions/NestedPointcutExpression.cs $
 * 
 * 9     5/11/06 12:44p Eaddy
 * o No longer uses a pointcut if its invalid
 * 
 * 8     5/05/06 3:37p Eaddy
 * o Added attributes so the debugger skips over methods
 * 
 * 7     4/03/06 12:23a Eaddy
 * o Now supports WithinCode and Within
 * 
 * 6     4/02/06 10:52p Eaddy
 * o Fixed parsing bug that causes exception to be thrown
 * 
 * 5     2/26/06 10:07p Eaddy
 * 
 * 4     2/26/06 9:12p Eaddy
 * o Now supports joinpoint matching
 * o Changed tabs to 3 spaces to conform with MS standard
 * 
 * 3     2/26/06 7:48p Eaddy
 * 
 * 2     2/24/06 8:17p Eaddy
 * o Initial checkin
 */

using System;
using System.Collections.Generic;

using Dbg = System.Diagnostics.Debug;

using Phx.Aop.Joinpoints;

namespace Phx.Aop.Pointcuts
{
   /// <summary>
   /// A pointcut expression containing one or more pointcut designators
   /// combined by logical operators (possibly nested)
   /// </summary>
   /// <remarks>
   /// We use parameter annotations for context exposure.  This means we
   /// do not need to support binding via expressions.  Expressions are
   /// only used for filtering join points.  Examples:
   /// 
   /// AspectJ:
   ///     [Pointcut("call(* *.*(int)) && args(i) && target(callee)")]
   ///     void someCall(int i, Foo callee) {}
   /// Wicca:
   ///     [Pointcut("call(* *.*(int))")]
   ///     void someCall([Arg(0)] int i, [Target] Foo callee) {}
   ///
   /// AspectJ:
   ///     [Pointcut("call(* *.*(int)) && target(Foo)")]
   ///     void someCall() {}
   /// Wicca:
   ///     [Pointcut("call(* *.*(int))")]
   ///     void someCall([Target] Foo /*callee*/) {}
   ///
   /// Note, to support negation we also must support this() and target().
   /// </remarks>
   /// <examples>
   /// <summary>
   /// A pointcut expression containing one or more pointcut designators
   /// combined by logical operators (possibly nested)
   /// </summary>
   /// <remarks>
   /// We use parameter annotations for context exposure.  This means we
   /// do not need to support binding via expressions.  Expressions are
   /// only used for filtering join points.  Examples:
   /// 
   /// AspectJ:
   ///     [Pointcut("call(* *.*(int)) && args(i) && target(callee)")]
   ///     void someCall(int i, Foo callee) {}
   /// Wicca:
   ///     [Pointcut("call(* *.*(int))")]
   ///     void someCall([Arg(0)] int i, [Target] Foo callee) {}
   ///
   /// AspectJ:
   ///     [Pointcut("call(* *.*(int)) && target(Foo)")]
   ///     void someCall() {}
   /// Wicca:
   ///     [Pointcut("call(* *.*(int))")]
   ///     void someCall([Target] Foo /*callee*/) {}
   ///
   /// Note, to support negation we also must support this() and target().
   /// <example>
   /// "!(call(* *.*(int)) && !target(Foo))"
   /// 
   ///                    PCE (ops: AND, trueValue: false)
   ///                     |
   ///         +-----------+----------+
   ///         |                      |
   ///        PCE                    PCE
   ///  pcd:call(* *.*(int))    pcd:target(Foo) (trueValue: false)
   /// </example>
   public class NestedPointcutExpression : PointcutExpression
   {
      static NestedPointcutExpression()
      {
         // If you don't register, the pointcut won't be found!
         CallPointcutExpression.register();
         ExecutionPointcutExpression.register();
         GetPointcutExpression.register();
         SetPointcutExpression.register();
         ThisPointcutExpression.register();
         TargetPointcutExpression.register();
         WithinPointcutExpression.register();
         WithinCodePointcutExpression.register();
      }

      BinaryPointcutExpression exprRoot = new BinaryPointcutExpression();

      enum ParseState { NORMAL, DESIGNATOR, NESTED };

      public override PointcutExpression parse(string expr)
      {
         BinaryPointcutExpression curr = exprRoot;
         BinaryPointcutExpression prevExpr = null;

         // Start position of the designator or parenthesized expression
         int start = -1;

         // Current state of the parser
         ParseState state = ParseState.NORMAL;

         // Number of unmatched open parens
         uint parens = 0;

         // Last character parsed
         char prev = ' ';

         // Is the parenthesized expression (not) negated?
         bool isPositive = true;

         // Optimization
         int len = expr.Length;

         // Optimization: Builds up the chars of the expression
         System.Text.StringBuilder sbSubExpr = new System.Text.StringBuilder();
         sbSubExpr.Length = len;
         int sbCount = 0;

         for (int i = 0; i < len; ++i)
         {
            char c = expr[i];

            if (c == '(')
            {
               ++parens;

               if (start == -1)
               {
                  start = i;
                  if (char.IsLetterOrDigit(prev))
                  {
                     state = ParseState.DESIGNATOR;
                  }
                  else
                  {
                     state = ParseState.NESTED;
                     isPositive = prev != '!';
                  }
               }
               else
               {
                  sbSubExpr[sbCount++] = c;
               }
            }
            else if (c == ')')
            {
               if (start == -1)
               {
                  System.Console.WriteLine("Pointcut declaration has too many close parens:");
                  System.Console.WriteLine(expr);
                  System.Console.WriteLine(string.Format("{{0,{0}}}", i + 1), '^');
                  return null;
               }

               Dbg.Assert(parens >= 0);

               if (--parens > 0)
               {
                  sbSubExpr[sbCount++] = c;
                  prev = c;
                  continue;
               }

               PointcutExpression newExpr = null;

               if (state == ParseState.NESTED)
               {
                  newExpr = new NestedPointcutExpression();
                  newExpr.trueValue = isPositive;
               }
               else if (state == ParseState.DESIGNATOR)
               {
                  int end = start;

                  // Back up to find the designator name (can also start with !)
                  while (--start >= 0 &&
                      (char.IsLetterOrDigit(expr[start]) ||
                      expr[start] == '!'))
                  { /* empty */ }

                  Dbg.Assert(end - start - 1 > 0);

                  // Extract the designator part: "!call(* *.*(*))" -> "!call"
                  string designator = expr.Substring(start + 1, end - start - 1);
                  newExpr = PointcutFactory.create(designator);
               }
               else
               {
                  Dbg.Assert(false); // Can't happen
               }

               if (newExpr != null)
               {
                  // Parse the expression part: "call(* *.*(*))" -> "* *.*(*)"
                  sbSubExpr.Length = sbCount;
                  if (newExpr.parse(sbSubExpr.ToString()) == null)
                     return null;
               }

               // Create a new expression leaf for the next expression
               curr.left = newExpr;
               curr.right = new BinaryPointcutExpression();
               prevExpr = curr;
               curr = curr.right;

               // Reset counters, state, and temp buffers
               start = -1;
               isPositive = true;
               sbCount = 0;
               sbSubExpr.Length = len - i;
            }
            else if (start == -1 &&
                ((c == '&' && prev == '&') ||
                (c == '|' && prev == '|')))
            {
               if (prevExpr == null || prevExpr.op != BinaryOperator.INVALID)
               {
                  System.Console.WriteLine(
                      "Invalid pointcut expression. Did not expect '{0}{0}' at this time:",
                      c);
                  System.Console.WriteLine(expr);
                  System.Console.WriteLine(string.Format("{{0,{0}}}", i + 1), "^^");
                  return null;
               }

               Dbg.Assert(curr != null);
               prevExpr.op = c == '&' ? BinaryOperator.AND : BinaryOperator.OR;
            }
            else if (start != -1)
            {
               sbSubExpr[sbCount++] = c;
            }

            prev = c;
         }

         if (start != -1)
         {
            System.Console.WriteLine("Pointcut declaration has too many open parens:");
            System.Console.WriteLine(expr);
            System.Console.WriteLine(string.Format("{{0,{0}}}", start + 1), '^');
            return null;
         }

         if (prevExpr != null)
         {
            Dbg.Assert(prevExpr.op == BinaryOperator.INVALID);
            prevExpr.right = null;
         }

         return this;
      }

      [System.Diagnostics.DebuggerStepThrough]
      public override bool innerMatches(string rhs)
      {
         return exprRoot.innerMatches(rhs);
      }

      [System.Diagnostics.DebuggerStepThrough]
      public override bool innerMatches(IJoinpoint rhs)
      {
         return exprRoot.innerMatches(rhs);
      }

      [System.Diagnostics.DebuggerStepThrough]
      public override bool innerMatches(PointcutExpression rhs)
      {
         return exprRoot.innerMatches(rhs);
      }

      public override string innerToString()
      {
         return exprRoot.ToString();
      }
   }
}
