//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Attributes/Handlers/AttributeHandlerBase.cs 4     5/24/06 9:53p Eaddy $
//
// Description:
//
//    Abstract base class for all extension attributes
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Attributes/Handlers/AttributeHandlerBase.cs $
 * 
 * 4     5/24/06 9:53p Eaddy
 * o Better diagnostics when the ctor param list for the inner
 * attribute doesn't match any ctor
 * 
 * 3     3/03/06 3:38p Eaddy
 * o Name change
 * 
 * 2     2/24/06 8:24p Eaddy
 * o Added checkin comment
 */

using System;

using Dbg = System.Diagnostics.Debug;

namespace Phx.Morph.Attributes
{
   /// <summary>
   /// Abstract base class for all extension attributes
   /// </summary>
   /// <remarks>
   /// Implements IExtensionAttribute to deal with initializing
   /// an extension attribute and storing its attached symbol
   /// and type name (e.g., System.Boolean).
   /// <br></br>
   /// An attribute can be created either by being instantiated
   /// from metadata, in which case the attribute will have an
   /// attached symbol, or by calling new, in which case the user
   /// must specify the name of the class the attribute should
   /// be considered attached to.
   /// </remarks>

   public abstract class AttributeHandlerBase : Attribute, IExtensionAttribute
   {
      Phx.Syms.Sym attachedSym = null;
      Phx.Types.Type attachedType = null;
      string attachedTypeString = null;
      string attachedName = null;

      /// <summary>
      /// Symbol of the attached 'thing' (assembly, class, field,
      /// method, etc.)
      /// </summary>
      /// <example>
      /// In this example, the extends attribute is attached to Bar
      /// and extends Foo:
      /// <code>
      ///   [Extends("Foo")]
      ///   class Bar { ... }
      /// </code>
      /// </example>

      public Phx.Syms.Sym AttachedSym
      {
         get { return this.attachedSym; }
      }

      public Phx.Types.Type AttachedType
      {
         get { return this.attachedType; }
      }

      /// <returns>
      /// The attached name or the name of the attached symbol if
      /// no attached name was specify
      /// </returns>

      public string Name
      {
         get
         {
            if (this.attachedName != null)
               return this.attachedName;
            else if (this.attachedSym != null)
               return this.attachedSym.NameString;
            else if (this.attachedType != null && this.attachedType.TypeSym != null)
               return this.attachedType.TypeSym.NameString;
            else
               return null;
         }
      }

      /// <summary>
      /// Use to initialize an attribute that was created using new
      /// </summary>
      /// <param name="attachedTypeString">
      /// Name of the type of the attached 'thing' (assembly, class, field,
      /// method, etc.)
      /// </param>
      /// <param name="attachedName"></param>

      public void
      Attach
      (
         string attachedTypeString,
         string attachedName
      )
      {
         // Verify args

         Dbg.Assert(attachedName != null && attachedName.Length != 0);
         Dbg.Assert(attachedTypeString != null && attachedTypeString.Length != 0);

         // Only call this method once

         Dbg.Assert(this.attachedTypeString == null &&
            this.attachedName == null);

         this.attachedTypeString = attachedTypeString;
         this.attachedName = attachedName;
      }

      /// <summary>
      /// Called after attribute creation to specify the
      /// 'thing' (attachedSym) that this attribute is attached
      /// to and to initialize the attribute
      /// </summary>
      /// <param name="attachedSym">
      /// The Phx symbol that this attribute is attached to
      /// or null if the attribute was created via the
      /// command-line</param>
      /// <param name="globalTypeTable">
      /// Type table used to resolve the type
      /// associated with the attribute
      /// </param>
      /// <returns>true on success</returns>

      virtual public bool
      Init
      (
         Phx.Syms.Sym attachedSym,
         Phx.Types.Table globalTypeTable
      )
      {
         // Sanity check: Only call once

         Dbg.Assert(this.attachedSym == null);

         // Step 1:  Resolve the field attachedType
         //
         //          If the AddField object comes from an attribute then
         //          attachedSym will be set and we can obtain the field attachedType
         //          to use.  If the AddField object is initialized via
         //          the command-line then we only have a string for the
         //          field attachedType and we have to resolve it to an actual
         //          Phoenix attachedType.

         if (attachedSym != null)
         {
            // We were created via an attribute. We have an attachedSym and
            // need the attachedType.

            this.attachedSym = attachedSym;
            this.attachedType = Helpers.GetSymType(attachedSym);
         }
         else if (this.attachedTypeString != null)
         {
            // We were created via the command-line. We have an type name
            // and need the attachedSym.

            if (this.attachedTypeString.Length == 0)
            {
               Helpers.WriteError("Field type is empty.");
               return false;
            }

            Phx.Types.Type findType = Helpers.FindAggrType(globalTypeTable,
               this.attachedTypeString, false);

            if (findType == null)
            {
               Helpers.WriteError("Invalid attribute type '{0}': " +
                  "Type could not be found (are you missing an assembly reference?).",
                  this.attachedTypeString);
               return false;
            }
            else if (findType.TypeSym == null)
            {
               Helpers.WriteError("Invalid attribute type '{0}': " +
                  "Type is not an aggregate type.",
                  this.attachedTypeString);
               return false;
            }

            this.attachedSym = null;

            this.attachedType = findType;
         }
         else
         {
            // Shouldn't happen

            Dbg.Assert(false);
         }

         Dbg.Assert(this.attachedType != null);
         Dbg.Assert(this.attachedSym != null || this.attachedName != null);

         return true;
      }

      /// <returns>
      /// true if the attached type is not null
      /// </returns>

      virtual public bool
      IsInited
      (
      )
      {
         return this.attachedType != null && this.Name != null;
      }

      public override string ToString()
      {
         return Name;
      }
   }
}
