//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Attributes/Handlers/ExtendsAttributeHandler.cs 5     5/24/06 9:53p Eaddy $
//
// Description:
//
//    Marks that the class contains extensions and specifies
//    which class its extending.  Any ifaces or base class
//    implemented by the attached class will be added
//    automatically.
//
// Example:
//
//       [Extends("Hello")]
//       class MyExt
//       {
//          [Add]
//          bool b;
//       }
//
// Remarks:
//
//    All extension classes must have this attribute attached
//    to the class.  When, later in the class definition, we
//    encounter an Add, Replace, etc. attribute we will find
//    the parent Extends attribute which will tell us which
//    class we are extending.
//
//    Handles adding:
//
//       Base classes
//       Base interfaces
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Attributes/Handlers/ExtendsAttributeHandler.cs $
 * 
 * 5     5/24/06 9:53p Eaddy
 * o Better diagnostics when the ctor param list for the inner
 * attribute doesn't match any ctor
 * 
 * 4     3/15/06 1:20p Eaddy
 * o Added ToString()
 * 
 * 3     3/03/06 3:38p Eaddy
 * o Name change
 * 
 * 2     2/24/06 8:23p Eaddy
 * o Now contains Attribute instead of deriving from it
 * o Uses Phx.Aop.dll which has generic AOP attributes
 * o Added checkin comment
 */

using System;

using System.Collections.Generic;

using Dbg = System.Diagnostics.Debug;

using Phx.Morph.Editors;

using Phx.Aop.Attributes;

namespace Phx.Morph.Attributes
{
   /// <summary>
   /// Marks that the class contains extensions and specifies
   /// which class its extending.  Any ifaces or base class
   /// implemented by the attached class will be added
   /// automatically.
   /// </summary>
   /// <remarks>
   /// All extension classes must have this attribute attached
   /// to the class.  When, later in the class definition, we
   /// encounter an Add, Replace, etc. attribute we will find
   /// the parent Extends attribute which will tell us which
   /// class we are extending.
   /// </remarks>

   public class ExtendsAttributeHandler : AttributeHandlerBase, IModuleMorphAttribute, IAttributeHandler
   {
      // ISSUE-BUG-t-marcea-2005/08/12-I'd like to support
      // specifying the System.Type instead of the FQN for
      // a class since it would allow us to catch more errors
      // at compile-time.  I had some problems, however. One
      // is that if I have one ctor that takes a string and one
      // that takes a System.Type, the compiler seems to pick
      // ctor(string) and implicitly calls System.Type.ToString().
      // Even after I removed ctor(string), when I try to
      // instantiate the attribute I get an exception that
      // Node_Visitor.dll couldn't be found.

      /// <summary>
      /// Fully qualified name of the class we will be morphing.
      /// </summary>
      /// <example>
      /// In this example, the extends attribute is attached to Bar
      /// and extends Foo:
      /// <code>
      ///   [Extends("Foo")]
      ///   class Bar { ... }
      /// </code>
      /// </example>

      ExtendsAttribute attr;

      /// <summary>
      /// Keeps track of all the instantiated ExtendsAttribute
      /// objects
      /// </summary>
      /// <remarks>
      /// This allows us to lookup an extends attribute
      /// later by specifying the FQN of the attached class
      /// </remarks>
      /// <seealso cref="GetExtendsAttribute"/>

      static List<ExtendsAttributeHandler> listInstantiatedAttributes =
         new List<ExtendsAttributeHandler>();

      public ExtendsAttributeHandler()
      { }

      public ExtendsAttributeHandler
      (
         string typeToExtendString
      )
      {
         this.attr = new ExtendsAttribute(typeToExtendString);
         listInstantiatedAttributes.Add(this);
      }

      public string AttributeName { get { return "Phx.Aop.Attributes.ExtendsAttribute"; } }

      public IExtensionAttribute Create(Phx.Syms.AttrSym attrSym, Type attrType, AttributeHelper helper)
      {
         ExtendsAttribute attr = helper.InstantiateAttribute(attrSym, attrType) as ExtendsAttribute;
         return new ExtendsAttributeHandler(attr.TypeToExtendString);
      }

      /// <summary>
      /// Allows us to lookup an already instantiated extends
      /// attribute based on the attached class
      /// </summary>
      /// <remarks>
      /// Sub-attributes of extends (Add, Replace, MorphExec, etc.)
      /// need an associated extends attribute to know which class
      /// they are manipulating.  This function allows them to lookup
      /// an already instantiated extends attribute.  The assumption
      /// is that attributes are created in the order they appear in
      /// the source code so if the class has an extends attribute
      /// it should have already been created.
      /// </remarks>
      /// <param name="ownerName">Name of the attached class</param>
      /// <returns>
      /// The extends attribute for the attached class (if any)
      /// </returns>

      static public ExtendsAttributeHandler GetExtendsAttribute(string ownerName)
      {
         // ISSUE-REVIEW-t-marcea-2005/08/12-Lock the shared data
         // structure just in case we're running multi-threaded

         lock (listInstantiatedAttributes)
         {
            foreach (ExtendsAttributeHandler extendsAttr in listInstantiatedAttributes)
            {
               if (extendsAttr.AttachedSym != null &&
                  extendsAttr.AttachedSym.NameString == ownerName)
               {
                  return extendsAttr;
               }
            }
         }

         return null;
      }

      /// <summary>
      /// Fully qualified name of the type this attribute
      /// extends
      /// </summary>

      public string TypeToExtendString
      {
         get { return this.attr.TypeToExtendString; }
         set { this.attr.TypeToExtendString = value; }
      }

      /// <summary>
      /// Called after attribute creation to specify the
      /// 'thing' (ownerSym) that this attribute is attached
      /// to and to initialize the attribute
      /// </summary>
      /// <param name="ownerSym">
      /// The Phx symbol that this attribute is attached to
      /// or null if the attribute was created via the
      /// command-line</param>
      /// <param name="globalTypeTable">
      /// Type table used to resolve the type
      /// associated with the attribute
      /// </param>
      /// <returns>true on success</returns>

      public override bool
      Init
      (
         Phx.Syms.Sym ownerSym,
         Phx.Types.Table globalTypeTable
      )
      {
         // Execute the base class initialization

         if (!base.Init(ownerSym, globalTypeTable))
         {
            return false;
         }

         // Verify params

         if (ownerSym == null &&
            (this.TypeToExtendString == null || this.TypeToExtendString.Length == 0))
         {
            Helpers.WriteError(
               "Failed to extend a class: Type to extend is null or empty.");
            return false;
         }

         Dbg.Assert(IsInited());

         return IsInited();
      }

      /// <returns>
      /// true if the AttachedSym is not null
      /// </returns>

      public override bool
      IsInited()
      {
         return this.TypeToExtendString != null && base.IsInited();
      }

      /// <summary>
      /// Morphs the type
      /// </summary>
      /// <param name="moduleEditor">Module where the type lives</param>
      /// <returns>true on success</returns>

      public bool
      Morph
      (
         ModuleEditor moduleEditor
      )
      {
         Dbg.Assert(IsInited());

         // ISSUE-REVIEW-t-marcea-2005/08/02-Not sure why
         // typeToExtendString can be null

         TypeEditor typeToExtend = null;
         if (this.TypeToExtendString != null)
         {
            typeToExtend = moduleEditor.FindType(this.TypeToExtendString);
            if (typeToExtend == null)
            {
               Helpers.WriteError("Failed to extend type '{0}'. Type '{0}' not found in the assembly '{1}'.",
                  this.TypeToExtendString,
                  moduleEditor.ToString());

               return false;
            }
         }

         TypeEditor ownerTypeEditor =
            new TypeEditor(this.AttachedType.AsAggrType);

         // Add implemented interfaces and base class

         int errors = 0;

         Phx.Types.BaseTypeLink baseType = ownerTypeEditor.AggrType.BaseTypeLinkList;
         while (baseType != null)
         {
            // We skip adding a base class for System.Object

            if (!Helpers.IsObjectType(baseType.BaseAggrType))
            {
               if (typeToExtend.AddBase(baseType) == null)
               {
                  ++errors;
               }
            }

            baseType = baseType.Next;
         }

         return errors == 0;
      }

      /// <returns>
      /// Type to extend
      /// </returns>
      public override string ToString()
      {
         return TypeToExtendString;
      }
   }
}
