//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// Description:
//
//    Helper class for modifying and querying a method
//
// Remarks:
//
//
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;

using Dbg = System.Diagnostics.Debug;

using Phx.Aop;
using Phx.Aop.Joinpoints;
using Phx.Aop.Attributes;

namespace Phx.Morph.Editors
{
   /// <summary>
   /// Helper class for modifying and querying a method
   /// </summary>

   public abstract class MethodEditor
   {
      protected TypeEditor typeEditor = null;
      protected internal Phx.Syms.FuncSym funcSym = null;
      protected static PhxJoinpoint jp = new PhxJoinpoint();

      /// <summary>
      /// Create a method editor using the type editor this method
      /// belongs to and the FuncSym of the method we want to edit
      /// </summary>

      public MethodEditor
      (
         TypeEditor typeEditor,
         Phx.Syms.FuncSym funcSym
      )
      {
         this.typeEditor = typeEditor;
         this.funcSym = funcSym;
      }

      /// <summary>
      /// Create a method editor using the FuncSym of the method
      /// we want to edit
      /// </summary>

      public MethodEditor
      (
         Phx.Syms.FuncSym funcSym
      )
      {
         if (funcSym.EnclosingAggrType != null)
            this.typeEditor = new TypeEditor(funcSym.EnclosingAggrType);

         this.funcSym = funcSym;
      }
      /// <summary>
      /// The FuncSym associated with this method
      /// </summary>

      public Phx.Syms.FuncSym FuncSym
      {
         get { return this.funcSym; }
      }

      /// <summary>
      /// Name of the method
      /// </summary>

      public Name Name
      {
         get { return this.funcSym.Name; }
      }

      /// <summary>
      /// Name of the method
      /// </summary>

      public string NameString
      {
         get { return this.funcSym.NameString; }
      }

      /// <summary>
      /// Fully qualified name of the method (Namespace.Type.Method)
      /// </summary>

      public override string
      ToString()
      {
         if (this.funcSym == null)
            return "(null)";
         else
            return this.typeEditor.ToString() + "." + this.funcSym.NameString + "()";
      }

      public bool IsAdvice
      {
         get
         {
            FuncSymAdviceExtObj adviceExtObj = FuncSymAdviceExtObj.Get(this.funcSym);
            return adviceExtObj == null ? false : adviceExtObj.isAdvice;
         }

         set
         {
            FuncSymAdviceExtObj adviceExtObj = FuncSymAdviceExtObj.Get(this.funcSym);
            if (adviceExtObj == null)
            {
               if (value == false)
               {
                  // Shortcut: If the extension object is not found then we
                  // assume the value is false. No need to create a new
                  // extension obj.

                  return;
               }

               adviceExtObj = new FuncSymAdviceExtObj();
               this.funcSym.AddExtensionObject(adviceExtObj);
            }

            adviceExtObj.isAdvice = value;
         }
      }

      public Phx.Syms.PropertySym PropertySym
      {
         get
         {
            Phx.Syms.PropertySymExtensionObject funcSymExtObj =
             Phx.Syms.PropertySymExtensionObject.GetExtensionObject(this.funcSym);

            if (funcSymExtObj == null)
            {
               return null;
            }
            else
            {
               return funcSymExtObj.PropertySym;
            }
         }

         set
         {
            Phx.Syms.PropertySymExtensionObject funcSymExtObj =
             Phx.Syms.PropertySymExtensionObject.GetExtensionObject(this.funcSym);

            if (funcSymExtObj == null)
            {
               if (value == null)
               {
                  // Shortcut: If the extension object is not found then we
                  // assume the value is null. No need to create a new
                  // extension obj.

                  return;
               }

               Phx.Lifetime lifetime = this.funcSym.Table.Unit.Lifetime;

               funcSymExtObj =
               Phx.Syms.PropertySymExtensionObject.New(lifetime);
               funcSymExtObj.PropertySym = value;
               this.funcSym.AddExtensionObject(funcSymExtObj);
            }
         }
      }

      /// <summary>
      /// Replaces the body of this method with the supplied method
      /// </summary>
      /// <param name="methodToCopy">Contains the new method body</param>
      /// <returns>true on success</returns>

      public bool
      ReplaceWith
      (
         MethodEditor methodToCopy
      )
      {
         Dbg.Assert(methodToCopy != null);
         if (methodToCopy == null)
         {
            return false;
         }

         if (!methodToCopy.CopyFuncUnit(this))
         {
            return false;
         }

         Helpers.WriteOutput("Replaced method: " + Name);

         Dbg.Assert(this.funcSym.FuncUnit != null);
         return this.funcSym.FuncUnit != null;
      }

      /// <summary>
      /// Tries to match the pointcuts against each instruction
      /// in the method body.  When a match occurs, indicating
      /// a joinpoint match, we weave the advice associated with
      /// the matching pointcut.
      /// </summary>
      /// <param name="pointcuts">Collection of Pointcut objects</param>
      /// <returns>number of errors</returns>

      public int
      Weave
      (
         List<Phx.Aop.PhxAdvice> advices
      )
      {
         Dbg.Assert(advices != null);
         if (advices == null)
            return 1;

         if (IsAdvice)
         {
            // Don't weave advice code

            return 0;
         }

         // ISSUE-TODO-t-marcea-2005/08/02-This is not really needed
         // anymore since we should run after the RaiseIR phase
         // (an assert would be more appropriate)

         RaiseToLir();

         FuncUnit funcUnit = this.funcSym.FuncUnit;
         Dbg.Assert(funcUnit != null);

         // Walk each instruction and try to match the pointcuts

         Phx.IR.Instr nextInstr = funcUnit.FirstInstr;

         int errors = 0;

         while (nextInstr != null)
         {
            Phx.IR.Instr instr = nextInstr;

            // ISSUE-REVIEW-t-marcea-2005/08/02-We don't want to
            // weave into advice. However, AspectJ allows weaving
            // advice into other advice.

            nextInstr = instr.Next;

            jp.Instr = instr;

            foreach (Phx.Aop.PhxAdvice advice in advices)
            {
               if (!advice.Enabled)
               {
                  continue;
               }

               if ((instr.Opcode == Phx.Common.Opcode.EnterFunc &&
                     advice.AdviceType == AdviceType.after) ||
                  (instr.Opcode == Phx.Targets.Archs.MSIL.Opcode.ret &&
                     advice.AdviceType == AdviceType.before))
               {
                  continue;
               }

               Dbg.Assert(advice.Pointcut != null);

               if (!advice.Pointcut.matches(jp))
               {
                  continue;
               }

               if (advice.PointcutExclusions != null &&
                  advice.PointcutExclusions.matches(jp))
               {
                  continue;
               }

               // Prevent redundant weaving
               if (advice.AlreadyApplied(jp))
               {
                  continue;
               }

               Phx.IR.Instr instrToInjectAfter;
               if (advice.AdviceType == AdviceType.before)
               {
                  if (instr.Opcode == Phx.Common.Opcode.EnterFunc)
                  {
                     instrToInjectAfter = instr;
                  }
                  else
                  {
                     instrToInjectAfter = instr.Prev;
                  }
               }
               else
               {
                  Dbg.Assert(advice.AdviceType == AdviceType.after);

                  instrToInjectAfter = instr;

                  uint origLine = Helpers.SafeGetSourceLine(instrToInjectAfter);
                  uint currLine = origLine;
                  while(currLine <= origLine && 
                     (instrToInjectAfter.Opcode != Phx.Targets.Archs.MSIL.Opcode.ret &&
                     !instrToInjectAfter.IsBranchInstr))
                  {
                     instrToInjectAfter = instrToInjectAfter.Next;
                     currLine = Helpers.SafeGetSourceLine(instrToInjectAfter);
                  }

                  if (instrToInjectAfter.Opcode == Phx.Targets.Archs.MSIL.Opcode.ret ||
                      instrToInjectAfter.IsBranchInstr)
                  {
                     instrToInjectAfter = instrToInjectAfter.Prev;
                  }
               }

               Helpers.WriteOutput("Applying advice '{0}' to '{1}'.",
                  advice, jp);

               bool success = Weave(instrToInjectAfter,
                                    advice.AdviceMethodWeaver,
                                    advice.AdviceType,
                                    jp);

               advice.AddAppliedJoinpoint(jp);

               if (!success)
               {
                  ++errors;
               }
            }
         }

         return errors;
      }

      /// <summary>
      /// Injects a method call into the method body
      /// </summary>
      /// <remarks>
      /// For calls to advice methods, the call may include special
      /// parameters which we need to pass it.
      /// These are indicated using special marker attributes (e.g.,
      /// (ArgAttribute) on the types of each parameter.  This
      /// requires us to load the arguments onto the stack before
      /// calling the method.  If any method arg cannot be satisfied,
      /// because it has an unknown or no marker attribute, we have
      /// to abort the call.
      /// </remarks>
      /// <param name="insertAfterInstr">
      /// Instruction to insert the method call after</param>
      /// <param name="methodToCall">Method to call</param>
      /// <param name="adviceType">Advice type that was matched</param>
      /// <param name="kind">Kind of pointcut that was matched</param>
      /// <param name="signature">Joinpoint signature that was matched</param>
      /// <returns>true if the call injection succeeded</returns>

      public abstract bool
      Weave
      (
         Phx.IR.Instr insertAfterInstr,
         MethodEditor methodToCall,
         Phx.Aop.AdviceType adviceType,
         Phx.Aop.Joinpoints.IJoinpoint jp
      );

      /// <summary>
      /// Adds this method as a MethodRef to the supplied module
      /// </summary>
      /// <param name="moduleDest">Module to add ourselves to</param>
      /// <returns>Method editor for the new MethodRef</returns>

      public MethodEditor
      CloneRef
      (
         ModuleEditor moduleDest
      )
      {
         // First clone the type ref

         TypeEditor typeDest = this.typeEditor.CloneRef(moduleDest);
         if (typeDest == null)
            return null;

         // Now clone the method ref

         Phx.Syms.FuncSym funcSymRef =
           moduleDest.AddFuncSymRef(this.funcSym);

         if (funcSymRef == null)
            return null;

         return MethodEditorFactory.CreateEditor(typeDest, funcSymRef);
      }

      /// <summary>
      /// Clones this method
      /// </summary>
      /// <remarks>
      /// Clones the function signature, symbol table, method body,
      /// and properties, importing types as necessary
      /// </remarks>
      /// <param name="classDest">Type to add the cloned method to</param>
      /// <param name="isOverride">
      /// True if this method should override a virtual method
      /// from its base class (equivalent to using the C# 'overrides'
      /// keyword).
      /// </param>
      /// <param name="methodAccess">
      /// Access rights for the new method
      /// </param>
      /// <returns>A method editor for the new method</returns>

      public MethodEditor
      Clone
      (
         TypeEditor classDest,
         bool isOverride,
         Phx.Syms.Access methodAccess
      )
      {
         // Verify arguments

         Dbg.Assert(classDest != null);
         if (classDest == null)
            return null;

         // Sanity checks

         Dbg.Assert(this.funcSym != null);
         Dbg.Assert(this.funcSym.Type != null);
         Dbg.Assert(this.funcSym.Type.IsFuncType);
         if (this.funcSym == null ||
           this.funcSym.Type == null ||
           !this.funcSym.Type.IsFuncType)
            return null;

         MethodEditor methodFind = classDest.FindMethod(funcSym);
         if (methodFind != null)
         {
            return methodFind;
         }

         Phx.Name newMethodName =
           Phx.Name.New(classDest.ModuleEditor.ModuleUnit.Lifetime,
            this.funcSym.Name);

         // Step 1:  Clone function signature/type

         // Must do a raise or we can't clone the function signature,
         // symbol table, etc.

         RaiseToLir();

         Phx.Types.FuncType newMethodType = CloneFuncSignature(classDest);
         Dbg.Assert(newMethodType != null);
         if (newMethodType == null)
            return null;

         // Step 2:  Create a symbol for the new function

         // GlobalDef indicates that we will be providing a definition
         // in this compilation moduleUnit

         Phx.Syms.FuncSym newMethodSym = Phx.Syms.FuncSym.New(classDest.ModuleEditor.SymTable,
                                                               this.funcSym.ExternId,
                                                               newMethodName,
                                                               newMethodType,
                                                               Phx.Syms.Visibility.GlobalDef);
         Dbg.Assert(newMethodSym != null);
         if (newMethodSym == null)
            return null;

         MethodEditor methodDest = MethodEditorFactory.CreateEditor(classDest, newMethodSym);
         Dbg.Assert(methodDest != null);
         if (methodDest == null)
            return null;

         // So we don't accidentally use it

         newMethodSym = null;

         // Step 3:  "Smart clone" the function moduleUnit

         CopyFuncUnit(methodDest);
         Dbg.Assert(methodDest.funcSym.FuncUnit != null);

         // ISSUE-TODO-t-marcea-2005/08/03-Step 4: Clone the attributes

         // Step 5:  Initialize/copy members
         CopyMembers(methodDest);

         // Override the access if requested
         if (methodAccess != Phx.Syms.Access._Illegal)
         {
            methodDest.funcSym.Access = methodAccess;
         }

         // ISSUE-REVIEW-t-marcea-2005/08/03-Instead of requiring
         // the user to specify isOverride, perhaps we should
         // automatically override.  This issue is that the user
         // can't specify the 'override' keyword since the method
         // is actually being overridden in a different class from
         // where the method definition comes from.

         if (isOverride)
         {
            Dbg.Assert(!methodDest.funcSym.IsAbstract);
            Dbg.Assert(!methodDest.funcSym.IsVirtual);
            methodDest.funcSym.IsVirtual = true;
         }

         // Step 6:  Add method to the class

         // ISSUE-REVIEW-t-marcea-2005/08/02-Why do we need to call
         // both AddMethod() and InsertInLexicalScope()?  One of these
         // should do the job of both since otherwise the type system
         // is inconsistent.

         classDest.AggrType.AddMethod(methodDest.funcSym);
         classDest.AggrType.TypeSym.InsertInLexicalScope(methodDest.funcSym, newMethodName);

         return methodDest;
      }

      /// <summary>
      /// Clones this method
      /// </summary>
      /// <remarks>
      /// Clones the function signature, symbol table, method body,
      /// and properties, importing types as necessary
      /// </remarks>
      /// <param name="classDest">Type to add the cloned method to</param>
      /// <returns>A method editor for the new method</returns>

      public MethodEditor
      Clone
      (
         TypeEditor classDest
      )
      {
         // Verify arguments

         Dbg.Assert(classDest != null);
         if (classDest == null)
            return null;

         return Clone(classDest, false, this.funcSym.Access);
      }

      /// <summary>
      /// Copies some members
      /// </summary>
      /// <param name="methodDest">Method to copy to</param>

      private void
      CopyMembers
      (
         MethodEditor methodDest
      )
      {
         // ISSUE-REVIEW-t-marcea-2005/08/03-My attempt at a clone
         // method. The following assignments appear here due to
         // trial and error.  I'm probably missing some assignments.

         methodDest.funcSym.IsNewSlot = this.funcSym.IsNewSlot;
         methodDest.funcSym.IsAbstract = this.funcSym.IsAbstract;
         methodDest.funcSym.IsVirtual = this.funcSym.IsVirtual;
         methodDest.funcSym.IsHideBySig = this.funcSym.IsHideBySig;
         methodDest.funcSym.Access = this.funcSym.Access;

         // ISSUE-HACK-t-marcea-2005/08/05-Lifetime is an issue here.

         if (this.FuncSym.FuncArgList != null)
         {
            Dbg.Assert(methodDest.funcSym.FuncArgList == null);

            // ISSUE-HACK-Weipingh-2005/10/10-Phx#2806 Really Copy over the 
            // FuncArgList. Make FuncArg Cloneable, and update owner when clone.

            Phx.Collections.FuncArgList.Iterator funcArgIter =
             new Phx.Collections.FuncArgList.Iterator(this.FuncSym.FuncArgList);

            while (funcArgIter.MoveNext())
            {
               Phx.Syms.FuncArg funcArg = funcArgIter.Curr;
               Phx.Syms.FuncArg newFuncArg = Phx.Syms.FuncArg.New(methodDest.funcSym,
                  funcArg.Name, funcArg.ParamId);

               newFuncArg.IsIn = funcArg.IsIn;
               newFuncArg.IsOut = funcArg.IsOut;
               newFuncArg.IsOptional = funcArg.IsOptional;
               newFuncArg.DefaultValueSym = funcArg.DefaultValueSym;

               // ISSUE-HACK-Weipingh-2005/10/10-Phx#2806 
               // Following code should strip Morpher related funcArg.CustomAttrSymList;
               // We don't really want copy all attributes over.
               // newFuncArg.MarshalSym = funcArg.MarshalSym;

               methodDest.funcSym.AppendFuncArg(newFuncArg);
            }
         }

         Phx.Syms.PropertySym propSymSrc = PropertySym;
         if (propSymSrc != null)
         {
            Dbg.Assert(this.funcSym.IsSpecialName && this.funcSym.IsHideBySig);

            Phx.Syms.PropertySym propSymDest =
             methodDest.typeEditor.FindProperty(propSymSrc.Name);
            Dbg.Assert(propSymDest != null,
               "Did you forget to add the Property first?");

            methodDest.PropertySym = propSymDest;

            if (propSymSrc.FuncSymGetter == this.funcSym)
            {
               propSymDest.FuncSymGetter = methodDest.funcSym;
            }
            else if (propSymSrc.FuncSymSetter == this.funcSym)
            {
               propSymDest.FuncSymSetter = methodDest.funcSym;
            }
            else
            {
               Helpers.WriteError(
                  "Unexpected method name '{0}' assocated with a Property: Expected get_X or set_X method name.",
                  this.NameString);
            }
         }

         // ISSUE-HACK-t-marcea-2005/08/03-Set the section to
         // use a section in the target module's moduleUnit, not
         // the source module's moduleUnit

         Phx.Syms.SectSym sectSym =
           Helpers.FindSectSym(methodDest.typeEditor.ModuleEditor.ModuleUnit.SymTable,
            ".text");
         Dbg.Assert(sectSym != null);

         methodDest.funcSym.AllocBaseSectSym = sectSym;
      }

      /// <summary>
      /// Clones the current method signature
      /// </summary>
      /// <param name="classDest">
      /// Class where the new method will reside
      /// </param>
      /// <returns></returns>

      private Phx.Types.FuncType
      CloneFuncSignature
      (
         TypeEditor classDest
      )
      {
         return CloneFuncSignature(classDest.ModuleEditor,
                                    classDest.AggrType,
                                    this.funcSym.FuncType);
      }

      /// <summary>
      /// Clones the current method's signature
      /// </summary>
      /// <remarks>
      /// The destination class is needed to remap the this
      /// ptr.
      /// </remarks>
      /// <param name="moduleEditor">
      /// Module where we will import the signature
      /// </param>
      /// <param name="classTypeDest">
      /// Class that will eventually have a method with the
      /// cloned signature
      /// </param>
      /// <param name="funcTypeToClone">
      /// Type for the function signature we want to clone
      /// </param>
      /// <returns>
      /// The new type for the cloned signature
      /// </returns>

      static public Phx.Types.FuncType
      CloneFuncSignature
      (
         ModuleEditor moduleEditor,
         Phx.Types.AggrType classTypeDest,
         Phx.Types.FuncType funcTypeToClone
      )
      {
         Dbg.Assert(moduleEditor != null);
         Dbg.Assert(funcTypeToClone != null);
         if (moduleEditor == null || funcTypeToClone == null)
            return null;

         // Step 1:  Create a Function Builder

         Phx.Types.FuncTypeBuilder funcTypeBuilder =
                 Phx.Types.FuncTypeBuilder.New(moduleEditor.ModuleUnit.TypeTable);

         Dbg.Assert(funcTypeBuilder != null);
         if (funcTypeBuilder == null)
            return null;

         funcTypeBuilder.Begin();

         // Step 2:  Clone the calling convention

         funcTypeBuilder.CallingConventionKind =
           funcTypeToClone.CallingConventionKind;

         // Step 3:  Import and clone the return type

         Phx.Types.Type newRetType =
           moduleEditor.AddTypeRef(funcTypeToClone.RetType);
         Dbg.Assert(newRetType != null);

         funcTypeBuilder.AppendRetType(newRetType);

         // Step 4:  Clone method args

         for (int i = 1; i <= funcTypeToClone.CountArgs(); ++i)
         {
            // ISSUE-REVIEW-t-marcea-2005/08/03-Do we need to change
            // this to support changing the type for Explicit this
            // pointer?

            Phx.Types.FuncArg argToCopy = funcTypeToClone.GetNthArgFuncArg(i);
            Dbg.Assert(argToCopy != null);

            Phx.Types.Type importedType;

            if (argToCopy.IsThisPtr && classTypeDest != null)
            {
               // Convert the this ptr type from the class we are importing
               // from to the class we are importing to

               Phx.Types.PtrType thisPtrToCopy = argToCopy.Type.AsPtrType;
               Dbg.Assert(thisPtrToCopy != null);
               if (thisPtrToCopy == null)
                  return null;

               importedType = Phx.Types.PtrType.New(moduleEditor.ModuleUnit.TypeTable,
                                                    thisPtrToCopy.PtrTypeKind,
                                                    thisPtrToCopy.BitSize,
                                                    classTypeDest,
                                                    null);
            }
            else
            {
               importedType = moduleEditor.AddTypeRef(argToCopy.Type);
            }

            Dbg.Assert(importedType != null);

            Phx.Types.FuncArg argCopy = Phx.Types.FuncArg.New(moduleEditor.ModuleUnit.TypeTable,
                                             argToCopy.FuncArgKind,
                                             importedType);
            Dbg.Assert(argCopy != null);

            funcTypeBuilder.AppendArgFuncArg(argCopy);
         }

         // Step 5:  Create the function type

         return funcTypeBuilder.GetFuncType();
      }

      /// <summary>
      /// Copies this method's symbol table
      /// </summary>
      /// <param name="methodDest">
      /// Method to copy the symbol table to
      /// </param>

      private void
      CopyFuncUnitSymTable
      (
         MethodEditor methodDest
      )
      {
         // Verify arguments

         Dbg.Assert(methodDest != null && methodDest.funcSym != null && methodDest.funcSym.FuncUnit != null);
         if (methodDest == null || methodDest.funcSym == null || methodDest.funcSym.FuncUnit == null)
            return;

         if (this.funcSym.FuncUnit == null ||
           this.funcSym.FuncUnit.SymTable == null)
            return;

         // Create the destination symbol table if needed

         bool isNewSymTable = false;

         Phx.Syms.Table symTableDest = methodDest.funcSym.FuncUnit.SymTable;
         if (symTableDest == null)
         {
            Phx.Syms.Table.New(methodDest.funcSym.FuncUnit, 2, true);
            symTableDest = methodDest.funcSym.FuncUnit.SymTable;
            isNewSymTable = true;
         }

         Dbg.Assert(symTableDest != null);

         // Sanity checks

         Dbg.Assert(this.funcSym.FuncUnit.SymTable.LocalIdMap != null);
         Dbg.Assert(this.funcSym.FuncUnit.SymTable.LocalIdMap.InternalMap != null);
         Dbg.Assert(symTableDest.Unit != null);
         Dbg.Assert(symTableDest.Unit.Lifetime != null);

         if (this.funcSym.FuncUnit.SymTable.LocalIdMap == null ||
           this.funcSym.FuncUnit.SymTable.LocalIdMap.InternalMap == null ||
           symTableDest.Unit == null ||
           symTableDest.Unit.Lifetime == null)
         {
            return;
         }

         // Reserve local id space ahead of time because insertion with a specific id
         // requires that the id be pre-allocated. If we don't insert with the same id
         // as the source symbol, we will not truly be cloning and this can cause
         // iteration problems as well as synchronization errors with the stack slots.

         symTableDest.LocalIdMap.AllocIds(this.funcSym.FuncUnit.SymTable.LocalIdMap.MaxId -
           symTableDest.LocalIdMap.MaxId);

         // Copy the symbol table

         foreach (Phx.Syms.Sym symToCopy in this.funcSym.FuncUnit.SymTable.LocalIdMap.InternalMap)
         {
            if (!isNewSymTable)
            {
               Phx.Syms.Sym symLookup =
               Helpers.GetNameMap(symTableDest).Lookup(symToCopy.Name);
            }

            Phx.Name name = Phx.Name.New(symTableDest.Unit.Lifetime,
                                       symToCopy.NameString);
            Dbg.Assert(name != null);

            if (symToCopy.IsLocalVarSym)
            {
               Phx.Syms.LocalVarSym varSymToCopy = symToCopy.AsLocalVarSym;
               Dbg.Assert(varSymToCopy != null);

               Phx.Types.Type importedType;

               if (varSymToCopy.IsThisParam)
               {
                  importedType = methodDest.typeEditor.AggrType;
               }
               else
               {
                  importedType = methodDest.typeEditor.ModuleEditor.AddTypeRef(varSymToCopy.Type);
               }

               Dbg.Assert(importedType != null);

               Phx.Syms.LocalVarSym symNew = Phx.Syms.LocalVarSym.New(symTableDest,
                  0,
                  name,
                  importedType,
                  varSymToCopy.StorageClass, varSymToCopy.LocalId);

               // ISSUE-REVIEW-t-marcea-2005/08/03-We employ a memberwise
               // clone here. Hopefully, it is safe.

               symNew.MemberwiseClone(varSymToCopy);

               // ISSUE-HACK-t-marcea-2005/08/08-Have to explicitly set
               // the location, MemberwiseClone() doesn't do this.

               Dbg.Assert(varSymToCopy.Loc != null);

               Phx.Syms.FrameLoc frameLocToCopy = varSymToCopy.Loc.AsFrameLoc;

               Dbg.Assert(frameLocToCopy != null);

               symNew.Loc = Phx.Syms.FrameLoc.New(symTableDest.Unit.Lifetime,
                  frameLocToCopy.ByteOffset);
            }
            else if (symToCopy.IsLabelSym)
            {
               Phx.Syms.Sym symNew = Phx.Syms.LabelSym.New(symTableDest,
                  symToCopy.ExternId, name);

               // ISSUE-REVIEW-t-marcea-2005/08/03-We employ a memberwise
               // clone here. Hopefully, it is safe.

               symNew.MemberwiseClone(symToCopy.AsLabelSym);
            }
            else
            {
               // How to handle other types of local symbols?

#if PHX_DEBUG_DUMPS
               //symToCopy.Dump();
#endif
               Dbg.Assert(false);
            }
         }
      }

      /// <summary>
      /// Raises this method to Low-level Intermediate Representation
      /// </summary>
      /// <remarks>
      /// Needed to get access to the local symbol table, function
      /// argument list, etc.
      /// </remarks>

      protected internal void
      RaiseToLir()
      {
         this.typeEditor.ModuleEditor.ModuleUnit.Raise(this.funcSym,
                        FuncUnit.LowLevelIRBeforeLayoutFuncUnitState);

         Dbg.Assert(this.funcSym.FuncUnit != null);
      }

      /// <summary>
      /// Raises this method to Mid-level Intermediate Representation
      /// </summary>
      /// <remarks>
      /// Not Yet Implemented. Needed to unstackify the method body.
      /// </remarks>

      private void
      RaiseToMir()
      {
         this.typeEditor.ModuleEditor.ModuleUnit.Raise(this.funcSym,
                        FuncUnit.SymbolicFuncUnitState);
         Dbg.Assert(this.funcSym.FuncUnit != null);
      }

      /// <summary>
      /// Copies this method's method body
      /// </summary>
      /// <param name="methodDest">Method we are copying to</param>
      /// <returns></returns>

      private bool
      CopyFuncUnit
      (
         MethodEditor methodDest
      )
      {
         Dbg.Assert(methodDest != null);
         Dbg.Assert(methodDest.typeEditor != null);
         if (methodDest == null || methodDest.typeEditor == null)
            return false;

         if (this.funcSym == null)
            return false;

         // Step 1:  Decode the IR so that we get a local symbol table

         RaiseToLir();

         FuncUnit funcUnitSrc = this.funcSym.FuncUnit;
         Dbg.Assert(funcUnitSrc != null);

         // Step 2:  Create a function moduleUnit for the new function

         // Save the old function moduleUnit (if any)

         Phx.FuncUnit oldFuncUnit = methodDest.funcSym.FuncUnit;

         Phx.FuncUnit funcUnitDest = Phx.FuncUnit.New(funcUnitSrc,
                                                      methodDest.FuncSym);

         Dbg.Assert(funcUnitDest != null);
         if (funcUnitDest == null)
            return false;

         methodDest.funcSym.FuncUnit = funcUnitDest;

         // Step 3:  Smart clone the symbol table

         CopyFuncUnitSymTable(methodDest);

         // ISSUE-HACK-t-marcea-2005/08/03-Need to hammer DebugInfo
         // to NULL or we get an assert in Phx.Debug.Info.New()

         Dbg.Assert(funcUnitDest.FuncSym != null);
         funcUnitDest.FuncSym.DebugInfo = null;

         // ISSUE-HACK-t-marcea-2005/08/03-We have to create debug
         // info or we get a NULL reference exception in
         // Phx.Targets.Runtime.VCCRT.Win.MSIL.Base.Encode()

         Phx.Debug.Info.New(funcUnitDest.Lifetime, funcUnitDest);

         funcUnitDest.DebugInfo.ZeroTagIsValid = true;

         // Step 4:  We have to null out Sym for Label instructions

         // ISSUE-HACK-t-marcea-2005/08/03-To workaround an assert in
         // IR.CopyRange(), we have to null out Sym for Label
         // instructions

         // Sanity checks

         Dbg.Assert(funcUnitSrc.FirstEnterInstr != null);
         Dbg.Assert(funcUnitSrc.FirstEnterInstr.Next != null);
         Dbg.Assert(funcUnitSrc.LastExitInstr != null);
         Dbg.Assert(funcUnitSrc.LastExitInstr.Prev != null);
         Dbg.Assert(funcUnitSrc.LastExitInstr.Prev.Prev != null);

         Phx.IR.Instr firstInstrToCopy = funcUnitSrc.FirstEnterInstr.Next;

         // ISSUE-HACK-t-marcea-2005/08/03-We purposely skip copying
         // the last instruction because IR.CopyRange() can't copy
         // ret instructions.

         Phx.IR.Instr lastInstrToCopy = funcUnitSrc.LastExitInstr.Prev.Prev;

         // Sanity check: We assume the last instruction is RET

#if PHX_DEBUG_SUPPORT
         Dbg.Assert(lastInstrToCopy.Next.IsBranchInstr &&
            lastInstrToCopy.Next.AsBranchInstr.OpcodeToString() == "ret");
#endif

         // Null out the Sym

         for (Phx.IR.Instr curr = firstInstrToCopy;
            curr != null && curr != lastInstrToCopy;
            curr = curr.Next)
         {
            if (curr.IsLabelInstr)
            {
               Dbg.Assert(curr.AsLabelInstr.Sym == null ||
               curr.AsLabelInstr.Sym.Name == null);
               curr.AsLabelInstr.Sym = null;
            }
         }

         // Step 5:  Setup enter, exit, and return instructions

         // Exit instruction

         Phx.IR.LabelInstr exitFuncInstr = Phx.IR.LabelInstr.New(funcUnitDest,
            Phx.Common.Opcode.ExitFunc,
            methodDest.funcSym);

         funcUnitDest.FirstInstr.InsertAfter(exitFuncInstr);

         // Enter instruction

         funcUnitDest.NewEnterFunc(methodDest.funcSym, exitFuncInstr);

         // ISSUE-HACK-t-marcea-2005/08/03-We must create our own RET
         // instruction because CopyRange() won't copy RET instructions
         // (same hack mentioned before)

         Phx.IR.BranchInstr retInstr = Phx.IR.BranchInstr.NewRet(funcUnitDest,
                                                   Phx.Targets.Archs.MSIL.Opcode.ret,
                                                   exitFuncInstr);

         // ISSUE-HACK-t-marcea-2005/08/12-

         funcUnitDest.Legalize.Instr(retInstr);

         // Step 6:  Copy the IR.  We skip the ENTER FUNC, END FUNC, and
         //          the last RET instruction

         Phx.IR.Instr lastInstr = null;
         Phx.IR.Instr firstInstr = Phx.IR.Instr.CopyRange(firstInstrToCopy,
                                                      lastInstrToCopy,
                                                      ref lastInstr);

         exitFuncInstr.InsertRangeBefore(firstInstr, lastInstr);
         lastInstr.InsertAfter(retInstr);

         // Step 7:  Redirect type and field references to use the target
         //          module instead of the source module

         bool success = methodDest.BackPatchImportedIR();
         if (!success)
         {
            return false;
         }

         // ISSUE-HACK-t-marcea-2005/08/12-

         methodDest.FuncUnit_ReaderPostProcess();

         if (oldFuncUnit != null)
         {
            // ISSUE-HACK-t-marcea-2005/08/03-Must explicitly remove the
            // old function (can't remember why but I know we need to).

            methodDest.typeEditor.ModuleEditor.ModuleUnit.RemoveChildUnit(oldFuncUnit);
         }

         // Step 9:  Add it to the module or else the function body won't
         //          be emitted

         // ISSUE-HACK-t-marcea-2005/08/03-Why isn't this determined
         // automatically?

         methodDest.typeEditor.ModuleEditor.ModuleUnit.AppendChildUnit(funcUnitDest);

         funcUnitDest.Context.PopUnit();

         return true;
      }

      /// <summary>
      /// Fix up the AliasTags
      /// </summary>
      /// <remarks>
      /// ISSUE-HACK-t-marcea-2005/08/12-FuncUnit.ReaderPostProcess()
      /// is protected so we were told to copy the entire method body.
      /// </remarks>

      private void
      FuncUnit_ReaderPostProcess()
      {
         // Now that the IR has been built up compute the initial set of
         // conservative aliasing information.

         Alias.Info aliasInfo = funcSym.FuncUnit.AliasInfo;
         Phx.Lifetime lifetime = funcSym.FuncUnit.Lifetime;

         // Allocate the reserved alias tag sets.

         aliasInfo.AllocateReservedTags();

         Alias.SymAnalyzer symAnalyzer = Alias.SymAnalyzer.New(
            lifetime, aliasInfo);

         // Assign default aliasing operands to intrinsics.

         symAnalyzer.Run(funcSym.FuncUnit);
         symAnalyzer.AnnotateFunc(funcSym.FuncUnit);
         symAnalyzer.Delete(funcSym.FuncUnit);
      }

      /// <summary>
      /// Fixup operands imported when we imported the function unit
      /// </summary>
      /// <remarks>
      /// The imported operands are associated with the old function
      /// unit's symbol table and old module's symbol table.  We have
      /// to create new operands that use the current method's and
      /// module's symbol table.  We modify the following operands:
      /// <list>
      /// <item>A symbol from the old module unit must be redirected
      /// to the new module unit</item>
      /// <item>If the imported method references symbols from the
      /// old module unit, redirect to new module unit symbols</item>
      /// <item>If an instance method is imported, redirect the this
      /// ptr to the class we imported to</item>
      /// <item>If the imported method references instance fields,
      /// redirect to use fields in the new class</item>
      /// <item>
      /// ISSUE-HACK-t-marcea-2005/08/08-Not consistent with how
      /// instance fields are redirected
      /// If the imported method references static fields, redirect
      /// to use field refs from the imported class</item>
      /// <item>If imported method references this ptr as local var,
      /// redirect to the new class</item>
      /// <item>Import all string constants</item>
      /// <item>If the imported method calls another function, import
      /// that function reference</item>
      /// </list>
      /// </remarks>

      private bool
      BackPatchImportedIR()
      {
         Phx.FuncUnit funcUnit = this.funcSym.FuncUnit;
         ModuleEditor moduleEditor = this.typeEditor.ModuleEditor;

         Phx.Syms.LocalVarSym thisPtrSym = null;

         Phx.IR.Instr firstInstr = funcUnit.FirstEnterInstr.Next;
         Dbg.Assert(firstInstr != null);
         if (firstInstr == null)
            return false;

         foreach (Phx.IR.Instr currInstr in Phx.IR.Instr.IterEditing(firstInstr))
         {
            // ISSUE-HACK-t-marcea-2005/08/12-IR.CopyRange() doesn't set
            // the funcUnit or clear the debug tag correctly so we have to
            // hammer these or else we'll get an IR consistency check error.

            currInstr.Unit = funcUnit;
            currInstr.DebugTag = 0;

            foreach (Phx.IR.Opnd currOpnd in Phx.IR.Opnd.IterEditingSrcAndDst(currInstr))
            {
               if (currOpnd.IsVarOpnd)
               {
                  // ISSUE-BUG-t-marcea-2005/08/03-Due to a bug in the
                  // PE Reader, local variables are made into MemOpnds
                  // instead of VarOpnds

                  Dbg.Assert(currOpnd.AsVarOpnd.Sym == null);
               }
               else if (currOpnd.IsImmOpnd)
               {
                  Phx.IR.ImmOpnd immOpnd = currOpnd.AsImmOpnd;
                  if (immOpnd.Sym != null)
                  {
                     // Trying to find a correspondance between a symbol from
                     // some other module unit to the current module unit.
                     // I use symbol name and symbol kind to match.

                     immOpnd.Sym =
                   Helpers.GetNameMap(moduleEditor.ModuleUnit.SymTable).Lookup(
                           immOpnd.Sym.Name, immOpnd.Sym.SymKind);

                     Dbg.Assert(immOpnd.Sym != null);
                  }
               }
               else if (currOpnd.IsMemOpnd)
               {
                  Phx.IR.MemOpnd memOpnd = currOpnd.AsMemOpnd;
                  if (memOpnd.Field != null && memOpnd.Field.FieldSym != null)
                  {
                     // ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
                     // for unmanaged implementation

                     bool hasThisAttribute = AttributeHelper.HasAttribute(
                        memOpnd.Field.FieldSym, typeof(ThisAttribute));

                     bool hasAddAttribute = AttributeHelper.HasAttribute(
                        memOpnd.Field.FieldSym, typeof(AddAttribute));

                     if (hasThisAttribute)
                     {
                        // Since this is a ldfld instruction, we know that "ldarg this"
                        // is the previous instruction.  By simply removing the ldfld
                        // instruction, we know that 'this' is on the stack which is
                        // what we want.
                        //
                        // Before              After
                        // ------              -----
                        // ldarg this          ldarg this
                        // ldfld [SR0] _this
                        //

                        Dbg.Assert(currInstr.Opcode == Phx.Targets.Archs.MSIL.Opcode.ldfld);
                        Dbg.Assert(currInstr.Prev.Opcode == Phx.Targets.Archs.MSIL.Opcode.ldarg);

                        currInstr.Remove();
                     }
                     else
                     {
                        Phx.Syms.FieldSym newFieldSym;

                        // Redirect instance and static fields

                        if (hasAddAttribute)
                        {
                           newFieldSym = this.typeEditor.FindField(
                              memOpnd.Field.FieldSym.Name);
                        }
                        else
                        {
                           newFieldSym = moduleEditor.AddFieldSymRef(
                              memOpnd.Field.FieldSym);
                        }

                        Dbg.Assert(newFieldSym != null);
                        if (newFieldSym == null)
                        {
                           return false;
                        }

                        // ISSUE-HACK-t-marcea-2005/08/03-Must hammer field
                        // to NULL to avoid assert in Phx.Types.Field.New()

                        newFieldSym.Field = null;

                        memOpnd.ResetField(Phx.Types.Field.New(newFieldSym.EnclosingAggrType,
                                                            newFieldSym.Type,
                                                            0,
                                                            newFieldSym));
                     }
                  }
                  else if (memOpnd.Sym != null)
                  {
                     if (memOpnd.Sym.IsLocalVarSym &&
                   memOpnd.Sym.AsLocalVarSym.IsThisParam)
                     {
                        if (thisPtrSym == null)
                        {
                           // Lazy lookup

                           thisPtrSym = Helpers.FindThisPtrLocalVar(
                              this.funcSym.FuncUnit.SymTable);
                        }

                        Dbg.Assert(thisPtrSym != null);

                        memOpnd.Sym = thisPtrSym;
                     }
                     else if (memOpnd.Sym.IsConstSym &&
                    Helpers.IsStringType(memOpnd.Sym.AsConstSym.Type))
                     {
                        Phx.Syms.ConstSym newConstSym =
                     moduleEditor.AddUserString(memOpnd.Sym.AsConstSym.Str);

                        if (newConstSym != null)
                        {
                           memOpnd.Sym = newConstSym;
                        }
                     }
                  }
               }
               else if (currOpnd.IsFuncOpnd)
               {
                  bool hasAddAttribute = AttributeHelper.HasAttribute(
                     currOpnd.AsFuncOpnd.FuncSym, typeof(AddAttribute));

                  Phx.Syms.FuncSym newFuncSym;

                  if (hasAddAttribute)
                  {
                     MethodEditor findMethod = this.typeEditor.FindMethod(
                        currOpnd.AsFuncOpnd.FuncSym);
                     Dbg.Assert(findMethod != null);

                     newFuncSym = findMethod.FuncSym;
                  }
                  else
                  {
                     newFuncSym = moduleEditor.AddFuncSymRef(
                        currOpnd.AsFuncOpnd.FuncSym);
                  }

                  Dbg.Assert(newFuncSym != null);
                  if (newFuncSym == null)
                     return false;

                  currOpnd.Instr.FuncSym = newFuncSym;
               }

               // ISSUE-HACK-t-marcea-2005/08/12-Must manually update
               // the alias tag as per ChrisMcK.

               if (currOpnd.IsDataflow)
               {
                  currOpnd.AliasTag = funcUnit.AliasInfo.GetTag(currOpnd);
               }
            }
         }

         return true;
      }

      public void
      SetProceed
      (
         MethodEditor proceedMethod
      )
      {
         Phx.FuncUnit funcUnit = this.funcSym.FuncUnit;
         ModuleEditor moduleEditor = this.typeEditor.ModuleEditor;

         Phx.IR.Instr firstInstr = funcUnit.FirstEnterInstr.Next;
         Dbg.Assert(firstInstr != null);
         if (firstInstr == null)
            return;

         foreach (Phx.IR.Instr currInstr in Phx.IR.Instr.IterEditing(firstInstr))
         {
            foreach (Phx.IR.Opnd currOpnd in Phx.IR.Opnd.IterEditingSrcAndDst(currInstr))
            {
               if (!currOpnd.IsFuncOpnd)
               {
                  continue;
               }

               Phx.Syms.FuncSym currFuncArg = currOpnd.AsFuncOpnd.FuncSym;
               Dbg.Assert(currFuncArg != null);

               currOpnd.Instr.FuncSym = proceedMethod.FuncSym;
            }
         }
      }
   }

   public class FuncSymAdviceExtObj : Phx.Syms.SymExtensionObject
   {
      private static uint id;
      public bool isAdvice = false;

      // Boilerplate code

      static FuncSymAdviceExtObj()
      {
         // Register

         FuncSymAdviceExtObj.id =
         Phx.ExtensionObject.Register(typeof(FuncSymAdviceExtObj).ToString());
      }

      public override uint Id
      {
         get { return FuncSymAdviceExtObj.id; }
         set { }
      }

      static public FuncSymAdviceExtObj
      Get
      (
         Phx.Syms.Sym sym
      )
      {
         return sym.FindExtensionObject(FuncSymAdviceExtObj.id)
            as FuncSymAdviceExtObj;
      }
   }
}
