//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Editors/ModuleEditor.cs 9     6/02/06 2:11a Eaddy $
//
// Description:
//
//    Helper class for modifying and querying a module
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Editors/ModuleEditor.cs $
 * 
 * 9     6/02/06 2:11a Eaddy
 * o Got rid of redundant verbosity flags
 * 
 * 8     5/24/06 9:53p Eaddy
 * o Better diagnostics when the ctor param list for the inner
 * attribute doesn't match any ctor
 * 
 * 7     5/24/06 8:56a Eaddy
 * o Changed file names but forgot to change class
 * names
 * 
 * 6     5/12/06 1:25a Eaddy
 * o Fixed bug where we had a double ptr
 * 
 * 5     3/15/06 1:17p Eaddy
 * o Added verbosity flag
 * 
 * 4     3/03/06 2:58p Asv2105
 * Refactored so all classes refer to base class of MethodWeaver rather
 * than MethodEditor, and create new instances through
 * MethodWeaverFactory.
 * 
 * 3     3/01/06 2:17a Eaddy
 * o Now uses Phx.Aop library for pointcuts, advice and joinpoint matching
 * 
 * 2     2/24/06 8:27p Eaddy
 * o Added checkin comment
 */

using System;

using Dbg = System.Diagnostics.Debug;

namespace Phx.Morph.Editors
{
   /// <summary>
   /// Helper class for modifying and querying a module
   /// </summary>

   public class ModuleEditor : System.Collections.IEnumerable
   {
      Phx.PEModuleUnit moduleUnit;

      public ModuleEditor
      (
         Phx.PEModuleUnit moduleUnit
      )
      {
         this.moduleUnit = moduleUnit;
      }

      /// <summary>
      /// Module unit associated with this module editor
      /// </summary>

      public Phx.PEModuleUnit ModuleUnit
      {
         get { return this.moduleUnit; }
      }

      /// <summary>
      /// Module unit's lifetime
      /// </summary>

      public Phx.Lifetime Lifetime
      {
         get { return this.moduleUnit != null ? this.moduleUnit.Lifetime : null; }
      }

      /// <summary>
      /// Module unit's symbol table
      /// </summary>

      public Phx.Syms.Table SymTable
      {
         get { return this.moduleUnit != null ? this.moduleUnit.SymTable : null; }
      }

      /// <summary>
      /// Allows you to enumerate over the types in the module
      /// </summary>

      public System.Collections.IEnumerator
      GetEnumerator
      (
      )
      {
         return new ModuleEditorEnumerator(this);
      }

      /// <returns>
      /// Name of the module
      /// </returns>

      public override string
      ToString()
      {
         if (this.moduleUnit == null)
            return "(null)";
         else
            return this.moduleUnit.NameString;
      }

      /// <param name="className">Type to find in this module</param>
      /// <returns>
      /// Type editor for the type
      /// </returns>

      public TypeEditor
      FindType
      (
         string className
      )
      {
         // Verify args

         Dbg.Assert(this.moduleUnit != null);
         Dbg.Assert(className != null && className.Length != 0);

         if (this.moduleUnit == null || className == null || className.Length == 0)
         {
            return null;
         }

         // Find the class

         Phx.Syms.MsilTypeSym classToExtendSym =
            Helpers.FindMsilTypeSym(this.moduleUnit.SymTable, className);

         if (classToExtendSym == null)
         {
            Helpers.WriteError("Failed to extend type '{0}'. Type '{0}' not found in the original assembly.",
               className);
            return null;
         }

         Dbg.Assert(classToExtendSym.Type != null);

         // Make sure the type is a class type

         Phx.Types.AggrType classToExtend = classToExtendSym.Type.AsAggrType;
         return classToExtend == null ? 
            null : 
            new TypeEditor(this, classToExtend);
      }

      /// <summary>
      /// Adds a reference to an assembly defined in another module,
      /// with the ability to rename it during import
      /// </summary>
      /// <remarks>
      /// If the type's assembly is not already referenced, this will
      /// add an [AssemblyRef] to the metadata.
      /// </remarks>
      /// <param name="importAssemblySym">Assembly sym to import</param>
      /// <param name="newAssemblyNameStr">
      /// Assembly name to use or null to use the imported assembly's
      /// name.  This is useful if you need to import "mscorlib" but
      /// call it "System".
      /// </param>
      /// <returns>The new reference assembly symbol</returns>
      public Phx.Syms.AssemblySym
      AddAssemblyRef
      (
         Phx.Syms.AssemblySym importAssemblySym,
         string newAssemblyNameStr
      )
      {
         // Verify args

         Dbg.Assert(this.moduleUnit != null && importAssemblySym != null);
         if (this.moduleUnit == null || importAssemblySym == null)
            return null;

         // See if the assembly has already been imported

         Phx.Name assemblyName = Phx.Name.New(Phx.GlobalData.GlobalLifetime,
            newAssemblyNameStr == null ? importAssemblySym.NameString : newAssemblyNameStr);

         Phx.Syms.AssemblySym assemblySym = FindAssemblySym(assemblyName);
         if (assemblySym != null)
         {
            return assemblySym;
         }
         else if (newAssemblyNameStr == null)
         {
            Helpers.WriteOutput("Importing assembly: " + assemblyName);

            // Add the Assembly to the symbol table so it gets
            // reflected in the metadata

            return Phx.Syms.AssemblySym.New(null,
               importAssemblySym.Manifest, assemblyName, this.moduleUnit.SymTable);
         }
         else
         {
            Helpers.WriteOutput("Importing assembly ref: " + assemblyName);

            // User requested a new name so we have to copy the manifest

            Phx.Manifest importManifest = importAssemblySym.Manifest;
            Phx.Manifest manifestCopy = Phx.Manifest.New(Phx.GlobalData.GlobalLifetime);
            manifestCopy.Name = assemblyName;
            manifestCopy.HashAlgorithm = importManifest.HashAlgorithm;
            manifestCopy.PublicKey = importManifest.PublicKey;
            manifestCopy.Ver = importManifest.Ver;
            manifestCopy.CultureStr = importManifest.CultureStr;

            return Phx.Syms.AssemblySym.New(null,
               manifestCopy, assemblyName, this.moduleUnit.SymTable);
         }
      }

      /// <summary>
      /// Adds a reference to an assembly defined in another module
      /// </summary>
      /// <remarks>
      /// If the type's assembly is not already referenced, this will
      /// add an [AssemblyRef] to the metadata.
      /// </remarks>
      /// <param name="importAssemblySym">Assembly sym to import</param>
      /// <returns>The new reference assembly symbol</returns>

      public Phx.Syms.AssemblySym
      AddAssemblyRef
      (
         Phx.Syms.AssemblySym importAssemblySym
      )
      {
         return AddAssemblyRef(importAssemblySym, null);
      }

      /// <summary>
      /// Adds a reference to a type defined in another module
      /// </summary>
      /// <remarks>
      /// If the type is not already referenced, this will add
      /// a [TypeRef] to the metadata. If the type's assembly
      /// is not already referenced, this will add an [AssemblyRef]
      /// to the metadata.
      /// </remarks>
      /// <param name="typeToImport">The type to import</param>
      /// <returns>A reference type for the imported type</returns>
      public Phx.Types.Type
      AddTypeRef
      (
         Phx.Types.Type typeToImport
      )
      {
         // Verify args

         Dbg.Assert(this.moduleUnit != null);
         Dbg.Assert(typeToImport != null);
         if (this.moduleUnit == null || typeToImport == null)
            return null;

         // Keep track of the original type

         Phx.Types.Type originalType = typeToImport;

         // Step 1:  See if the type's symbol is already in the symbol table

         // Dereference pointer types

         while (typeToImport.IsPtrType)
         {
            typeToImport = typeToImport.AsPtrType.ReferentType;
            Dbg.Assert(typeToImport != null);
         }

         // Shortcut: If the type is a primitive type then we don't
         // need to import it since Phoenix recognizes these by default.

         if (typeToImport.IsPrimType ||
            (typeToImport.IsAggrType &&
            typeToImport.AsAggrType.IsRuntimeBuiltInType))
         {
            return typeToImport;
         }

         Dbg.Assert(typeToImport.TypeSym != null);

         if (typeToImport.TypeSym.Name.IsNull)
         {
            // If the name is null then the type is not a user-defined
            // type, its an internal Phx type.  Just return it.

            return originalType;
         }

         // See if the type is already in the symbol table

         Phx.Syms.MsilTypeSym typeSym = Helpers.FindMsilTypeSym(this.moduleUnit.SymTable, typeToImport.TypeSym.Name);
         if (typeSym != null)
         {
            // Symbol has already been imported

            return typeSym.Type;
         }

         // Step 2:  Import the assembly if needed

         Dbg.Assert(typeToImport.TypeSym != null);
         Dbg.Assert(typeToImport.TypeSym.ResolutionScopeSym != null);

         // Get the assembly that defines the type

         Phx.Syms.AssemblySym importAssemblySym = typeToImport.TypeSym.ResolutionScopeSym.AsAssemblySym;
         if (importAssemblySym == null)
         {
            importAssemblySym = typeToImport.TypeSym.ResolutionScopeSym.LexicalParentSym.AsAssemblySym;
         }

         Dbg.Assert(importAssemblySym != null);

         Phx.Syms.AssemblySym assemblySym = AddAssemblyRef(importAssemblySym);

         Dbg.Assert(assemblySym != null);
         if (assemblySym == null)
            return null;

         // Step 3:  Create a symbol for the new type

         Phx.Name newTypeName = Phx.Name.New(this.moduleUnit.Lifetime,
            typeToImport.TypeSym.NameString);

         Phx.Syms.MsilTypeSym newTypeSym;

         newTypeSym = Phx.Syms.MsilTypeSym.New(this.moduleUnit.SymTable,
            newTypeName, 0);

         // Step 3:  Insert the new type as a child of the assembly

         assemblySym.InsertInLexicalScope(newTypeSym, newTypeName);

         // Step 4:  Create a new reference type that will contain the 
         //          type ref and any other members we reference

         Phx.Types.AggrType newRefType = Phx.Types.AggrType.NewDynamicSize(this.moduleUnit.TypeTable,
            newTypeSym);

         // The symbol table may have both a type ref and a type def
         // for the same type.  However, the type ref version doesn't
         // have most of the type info (e.g. IsSelfDescribing). Try
         // to find the high-fidelity type def version.

         // ISSUE-REVIEW-t-marcea-2005/08/03-Its confusing that a
         // TypeSym may or may not have some of its members initialized.
         // We need TypeSymRef and TypeSymDef to eliminate confusion.

         Phx.Types.AggrType typeToImportDef = Helpers.FindAggrType(
            this.moduleUnit.TypeTable,
            typeToImport.TypeSym.Name,
            true);

         if (typeToImportDef != null)
         {
            Dbg.Assert(typeToImportDef.IsDef);

            // Determines if it shows up in the metadata as "Class"
            // or "ValueClass"

            newRefType.IsSelfDescribing = typeToImportDef.IsSelfDescribing;
         }
         else
         {
            // Will show up as "Class" not "ValueClass"

            // ISSUE-HACK-t-marcea-2005/08/02-While this is a good default
            // since most types are classes not value types (structs), the
            // right solution would be chase down the type ref to get to
            // the type def.  However, this would involve being able to
            // locate and load the assembly associated with a type ref,
            // which is what Fusion does (read: not bloody likely).

            newRefType.IsSelfDescribing = true;
         }

         // Specify TypeRef not TypeDef

         newRefType.IsDef = false;

         Helpers.WriteOutput("Importing type ref: " + 
            Helpers.SafeGetName(newRefType));

         return newRefType;
      }

      /// <summary>
      /// Adds a reference to a string constant
      /// </summary>
      /// <remarks>
      /// Adds a User String to the metadata if its not already present
      /// </remarks>
      /// <param name="s"></param>
      /// <returns></returns>

      public Phx.Syms.ConstSym
      AddUserString
      (
         string s
      )
      {
         // Verify args

         Dbg.Assert(this.moduleUnit != null);
         Dbg.Assert(s != null);
         if (this.moduleUnit == null || s == null)
            return null;

         // See if the string already exists

         Phx.Syms.ConstSym foundSym = Helpers.FindOrCreateUserString(this.moduleUnit.SymTable, 
                                                                     this.moduleUnit,
                                                                     s);
         Dbg.Assert(foundSym != null);

         Helpers.WriteOutput("Adding user string: \"" + s + "\".");

         return foundSym;
      }

      /// <summary>
      /// Adds a reference to a method that is defined in another
      /// module
      /// </summary>
      /// <remarks>
      /// Adds a [MemberRef] to the metadata if its not already
      /// present.  If the type of the function is not already
      /// referenced, this will add a [TypeRef] to the
      /// metadata. If the type's assembly is not already referenced,
      /// this will add an [AssemblyRef] to the metadata.
      /// </remarks>
      /// <param name="symToImport">Function symbol to import</param>
      /// <returns>A reference FuncSym for the imported function</returns>

      public Phx.Syms.FuncSym
      AddFuncSymRef
      (
         Phx.Syms.FuncSym symToImport
      )
      {
         // Verify args

         Dbg.Assert(this.moduleUnit != null);
         Dbg.Assert(symToImport != null);
         if (this.moduleUnit == null || symToImport == null)
            return null;

         // Sanity checks

         Dbg.Assert(symToImport.Type != null);

         // Step 1:  Import the enclosing type (if needed)

         Phx.Types.Type typeToImport = symToImport.AsFuncSym.EnclosingAggrType;
         Dbg.Assert(typeToImport != null);
         if (typeToImport == null)
            return null;

         // Verify that we can create a method ref

         if (typeToImport.AsAggrType.IsDef && 
            typeToImport.AsAggrType.Access != Phx.Syms.Access.Public)
         {
            Helpers.WriteError("Cannot import method reference '{0}.{1}': " +
               "The type '{0}' is inaccessible (Access is '{2}' but should be '{3}').",
               typeToImport.AsAggrType.TypeSym.NameString,
               symToImport.NameString,
               typeToImport.AsAggrType.Access,
               Phx.Syms.Access.Public);
            return null;
         }
         else if (symToImport.IsDef &&
            symToImport.Access != Phx.Syms.Access.Public)
         {
            Helpers.WriteError("Cannot import method reference '{0}.{1}': " +
               "The method is inaccessible (Access is '{2}' but should be '{3}').",
               typeToImport.AsAggrType.TypeSym.NameString,
               symToImport.NameString,
               symToImport.Access,
               Phx.Syms.Access.Public);
            return null;
         }

         Phx.Types.Type importedType = AddTypeRef(typeToImport);

         Dbg.Assert(importedType != null);
         Dbg.Assert(importedType.IsAggrType);
         Dbg.Assert(importedType.TypeSym != null);
         if (importedType == null || !importedType.IsAggrType ||
            importedType.TypeSym == null)
         {
            return null;
         }

         // Step 2:  Import the function

         // See if the function already exists

         Phx.Syms.FuncSym foundSym = Helpers.FindMethod(importedType.AsAggrType, 
            symToImport);

         if (foundSym != null)
         {
            return foundSym;
         }

         // Step 3:  Create a new function symbol

         Phx.Name newSymName = Phx.Name.New(this.moduleUnit.Lifetime,
            symToImport.NameString);

         // Step 4:  Clone the function signature. This is needed
         //          so that all the types in the signature are
         //          imported correctly.

         Phx.Types.FuncType newFuncType = MethodEditor.CloneFuncSignature(this,
            null,
            symToImport.Type.AsFuncType);

         Phx.Syms.FuncSym newFuncSym =
            Phx.Syms.FuncSym.New(this.moduleUnit.SymTable,
               symToImport.AsFuncSym.ExternId,
               newSymName,
               newFuncType,
               Phx.Syms.Visibility.GlobalRef);

         // Its a MethodRef not a MethodDef

         // ISSUE-REVIEW-t-marcea-2005/08/02-Why do we have both
         // FuncSym.Visibility.GlobalDef and AggrType.IsDef?
         // How are they related?

         newFuncSym.Visibility = Phx.Syms.Visibility.GlobalRef;

         // Step 5:  Add the new method to the class

         // ISSUE-REVIEW-t-marcea-2005/08/02-Why do we need to call
         // both AddMethod() and InsertInLexicalScope()?  One of these
         // should do the job of both since otherwise the type system
         // is inconsistent.

         importedType.AsAggrType.AddMethod(newFuncSym);
         importedType.TypeSym.InsertInLexicalScope(newFuncSym, newSymName);

         Helpers.WriteOutput("Importing method ref: {0}.{1}().",
            typeToImport.AsAggrType.TypeSym.NameString,
            symToImport.NameString);

         return newFuncSym;
      }

      /// <summary>
      /// Adds a reference to a field that is defined in another module
      /// </summary>
      /// <remarks>
      /// Adds a [MemberRef] or User String to the metadata if its
      /// not already present.  If the type of the field/function/etc
      /// is not already referenced, this will add a [TypeRef] to the
      /// metadata. If the type's assembly is not already referenced,
      /// this will add an [AssemblyRef] to the metadata.
      /// </remarks>
      /// <param name="symToImport">Field to import</param>
      /// <returns>A reference FieldSym for the imported field</returns>

      public Phx.Syms.FieldSym
      AddFieldSymRef
      (
         Phx.Syms.FieldSym symToImport
      )
      {
         // Verify args

         Dbg.Assert(this.moduleUnit != null);
         Dbg.Assert(symToImport != null);
         if (this.moduleUnit == null || symToImport == null)
            return null;

         // Sanity checks

         Dbg.Assert(symToImport.Type != null);

         // Step 1:  Import the enclosing type (if needed)

         Phx.Types.Type typeToImport = symToImport.AsFieldSym.EnclosingAggrType;
         Dbg.Assert(typeToImport != null);
         if (typeToImport == null)
            return null;

         // Verify that we can create a field ref

         if (typeToImport.AsAggrType.IsDef &&
            typeToImport.AsAggrType.Access != Phx.Syms.Access.Public)
         {
            Helpers.WriteError("Cannot import field reference '{0}.{1}': " +
               "The type '{0}' is inaccessible (Access is '{2}' but should be '{3}').",
               typeToImport.AsAggrType.TypeSym.NameString,
               symToImport.NameString,
               typeToImport.AsAggrType.Access,
               Phx.Syms.Access.Public);
            return null;
         }
         else if (symToImport.IsDef && 
            symToImport.Access != Phx.Syms.Access.Public)
         {
            Helpers.WriteError("Cannot import field reference '{0}.{1}': " +
               "The field is inaccessible (Access is '{2}' but should be '{3}').",
               typeToImport.AsAggrType.TypeSym.NameString,
               symToImport.NameString,
               symToImport.Access,
               Phx.Syms.Access.Public);
            return null;
         }

         Phx.Types.Type importedType = AddTypeRef(typeToImport);

         Dbg.Assert(importedType != null);
         Dbg.Assert(importedType.IsAggrType);
         Dbg.Assert(importedType.TypeSym != null);
         if (importedType == null || !importedType.IsAggrType ||
            importedType.TypeSym == null)
         {
            return null;
         }

         // Step 2:  Import the field

         // See if the field already exists

         Phx.Syms.FieldSym foundSym = Helpers.FindField(importedType.AsAggrType, symToImport.Name);
         if (foundSym != null)
         {
            return foundSym;
         }

         // Step3:   Create a new field symbol

         Phx.Name newSymName = Phx.Name.New(this.moduleUnit.Lifetime,
            symToImport.NameString);

         Phx.Syms.FieldSym newFieldSym = Phx.Syms.FieldSym.New(this.moduleUnit.SymTable,
            symToImport.AsFieldSym.ExternId,
            newSymName, symToImport.Type);

         // ISSUE-HACK-t-marcea-2005/08/02-Phx should know if NeedsFixup
         // needs to occur. Currently we have to hammer this field.

         newFieldSym.NeedsFixup = true;

         // Step 4:  Add the field to the class

         // ISSUE-REVIEW-t-marcea-2005/08/02-Why do we need to call
         // both AddField() and InsertInLexicalScope()?  One of these
         // should do the job of both since otherwise the type system
         // is inconsistent.  I'm assuming we don't use AppendFieldSym()
         // because the type is a TypeRef not a TypeDef (I assert this
         // below).

         Dbg.Assert(!importedType.AsAggrType.IsDef);

         importedType.AsAggrType.AddField(newFieldSym);
         importedType.TypeSym.InsertInLexicalScope(newFieldSym, newSymName);

         Helpers.WriteOutput("Importing field ref: " +
            symToImport.NameString);

         return newFieldSym;
      }

      /// <summary>
      /// Checks to see if the assembly is in this module's symbol table
      /// </summary>
      /// <param name="assemblyName">Assembly to look for</param>
      /// <returns>True if the assembly is found</returns>

      public Phx.Syms.AssemblySym
      FindAssemblySym
      (
         Phx.Name assemblyName
      )
      {
         Dbg.Assert(this.moduleUnit != null && assemblyName != null);
         if (this.moduleUnit == null || assemblyName == null)
            return null;

         Phx.Syms.NameMap nameMap = Helpers.GetNameMap(this.moduleUnit.SymTable);
         if (nameMap == null)
            return null;

         // Note that there might be a number of symbols with
         // identical names, so search through until we have an
         // assembly reference.

         Phx.Syms.Sym found = nameMap.Lookup(assemblyName);
         while (found != null)
         {
            if (found.IsAssemblySym)
            {
               return found.AsAssemblySym;
            }

            found = nameMap.LookupNext(found);
         }

         return null;
      }

      /// <summary>
      /// Checks to see if the assembly is in this module's symbol table
      /// </summary>
      /// <param name="assemblyName">Assembly to look for</param>
      /// <returns>True if the assembly is found</returns>

      public Phx.Syms.AssemblySym
      FindAssemblySym
      (
         string assemblyName
      )
      {
         return FindAssemblySym(Phx.Name.New(Phx.GlobalData.GlobalLifetime, assemblyName));
      }
   }

   public class ModuleEditorEnumerator : System.Collections.IEnumerator
   {
      Phx.Collections.IdToSymMap.Iterator iter;
      ModuleEditor moduleEditor = null;

      public ModuleEditorEnumerator(ModuleEditor moduleEditor)
      {
         this.moduleEditor = moduleEditor;

         Phx.PEModuleUnit unit = moduleEditor.ModuleUnit;

         if (unit != null &&
            unit.SymTable != null &&
            unit.SymTable.LocalIdMap != null &&
            unit.SymTable.LocalIdMap.InternalMap != null)
         {
            this.iter = new Phx.Collections.IdToSymMap.Iterator(unit.SymTable.LocalIdMap.InternalMap);
         }
      }

      public object Current
      {
         get
         {
            if (this.iter.Curr != null)
            {
               Dbg.Assert(this.iter.Curr.IsMsilTypeSym);

               return new TypeEditor(this.moduleEditor, 
                  this.iter.Curr.AsMsilTypeSym.Type.AsAggrType);
            }
            else
            {
               Dbg.Fail("Forget to call MoveNext()?");
               return null;
            }
         }
      }

      public bool MoveNext()
      {
         while (this.iter.MoveNext())
         {
            if (this.iter.Curr.IsMsilTypeSym)
               return true;
         }

         return false;
      }

      public void Reset()
      {
         this.iter.Reset();
      }
   }
}
