//---------------------------------------------------------------------
//  This file is part of the CLR Managed Debugger (mdbg) Sample.
// 
//  Copyright (C) Microsoft Corporation.  All rights reserved.
//---------------------------------------------------------------------
using System;
using System.IO;
using System.Collections;
using System.Diagnostics;
using System.Threading;
using System.Reflection;
using System.Text;
using System.Security.Permissions;
using System.Globalization;

using Microsoft.Samples.Debugging.MdbgEngine;
using Microsoft.Samples.Debugging.CorDebug;
using Microsoft.Samples.Debugging.CorDebug.NativeApi;

namespace Microsoft.Samples.Tools.Mdbg
{

    /// <summary>
    /// Summary description for Class1.
    /// </summary>
    public class MDbgShell: IMDbgShell
    {

        public IMDbgIO IO
        {
            get
            {
                return m_io;
            }
            set
            {
                Debug.Assert(value!=null);
                m_io = value;
            }
        }

        public MDbgEngine Debugger
        {
            get
            {
                return m_debugger;
            }
            set
            {
                if(value==null)
                {
                    throw new ArgumentException();
                }
                m_debugger = value;
            }
        }

        public int Verbosity
        {
            get { return m_verbosity; }
            set { m_verbosity = value; }
        }

        public IMDbgCommandCollection Commands
        {
            get
            {
                return m_commands;
            }
        }

        // Quit w/ an exit code.
        public void QuitWithExitCode(int exitCode)
        {
            m_run = false;
            m_exitCode = exitCode;
        }        

        public IMDbgFileLocator FileLocator
        {
            get
            {
                return m_fileLocator;
            }
        }

        public IMDbgSourceFileMgr  SourceFileMgr
        {
            get
            {
                return m_sourceFileMgr;
            }
        }

        public IDictionary Properties
        {
            get
            {
                return m_properties;
            }
        }

        IBreakpointParser m_BreakpointParser;
        
        /// <summary>
        /// Get the default Breakpoint parser for this collection.
        /// </summary>
        /// <remarks> The breakpoint collection maintains a default breakpoint parser. 
        /// Extensions can get the parser so that they can use the share the parsing implementation to get the 
        /// same breakpoint syntax as the rest of the shell. This encourages a uniform breakpoint syntax.
        /// Extensions can also set the parser so that they can override and even extend the breakpoint syntax.
        /// This may be null (and is in fact null by default), though it is reasonable for extensions to 
        /// expect that the shell supplies a parser.
        /// </remarks>
        public IBreakpointParser BreakpointParser
        {
            get 
            {
                return m_BreakpointParser;
            }
            set 
            {
                m_BreakpointParser = value;
            }
        }
        
        IExpressionParser m_ExpressionParser;
        
        public IExpressionParser ExpressionParser
        {
            get 
            {
                return m_ExpressionParser;
            }
            set
            {
                m_ExpressionParser = value;
            }
        }


        public event CommandExecutedEventHandler OnCommandExecuted;

        public int Start(string[] commandLineArguments)
        {
            this.Init(commandLineArguments);

            if (!m_nologo) PrintStartupLogo();
            return this.RunInputLoop();
        }

        //////////////////////////////////////////////////////////////////////////////////
        //
        // Customization methods (to be overriden in aditional skins).
        //
        //////////////////////////////////////////////////////////////////////////////////

        protected virtual void Init(string[] commandLineArguments)
        {
            string[] initialCommands = null;

            // process startup commands
            if( commandLineArguments.Length != 0 )
            {
                int i = 0;

                ArrayList startupCommands = new ArrayList();

                while(  i < commandLineArguments.Length &&
                        commandLineArguments[i].Length > 1 && 
                        (commandLineArguments[i][0] == '-' || commandLineArguments[i][0] == '/'))
                {
                    // Command-line switches
                    StringBuilder sb = new StringBuilder();
                    Debug.Assert(commandLineArguments[i][0] == '-' || commandLineArguments[i][0] == '/');
                    sb.Append(commandLineArguments[i].Substring(1));
                    ++i;

                    ConsoleColor fc = Console.ForegroundColor;
                    Console.ForegroundColor = ConsoleColor.Red;

                    switch(sb.ToString().ToLower())
                    {
                        case "nologo": 
                            m_nologo = true; 
                            break;

                        case "verbosity":

                            if (i >= commandLineArguments.Length)
                            {
                                Console.WriteLine("MDbg: Must specify the verbosity level: 0, 1, or 2.");
                            }
                            else if (!int.TryParse(commandLineArguments[i], out m_verbosity) ||
                                m_verbosity < 0 ||
                                m_verbosity > 3)
                            {
                                Console.WriteLine("MDbg: Invalid verbosity '" + commandLineArguments[i] + 
                                               "' Expected 0 (no msgs), 1 (errors only), 2 (errors + warnings), or 3 (all msgs).");
                            }

                            ++i; // Skip over the verbosity
                            break;

                        default:
                            Console.WriteLine("MDbg: Unknown command-line argument '" + commandLineArguments[i] + "'.");
                            break;
                    }

                    Console.ForegroundColor = fc;
                }

                if (i < commandLineArguments.Length &&
                    commandLineArguments[i].Length > 1 && commandLineArguments[i][0] == '!')
                {
                    // ! commands on command line
                    while(i<commandLineArguments.Length)
                    {
                        StringBuilder sb = new StringBuilder();
                        Debug.Assert(commandLineArguments[i][0]=='!');
                        sb.Append(commandLineArguments[i].Substring(1));
                        ++i;
                        while(i<commandLineArguments.Length &&
                              !(commandLineArguments[i].Length>1 && commandLineArguments[i][0]=='!'))
                        {
                            sb.Append(' ');
                            sb.Append(commandLineArguments[i]);
                            ++i;
                        }
                        startupCommands.Add(sb.ToString());
                    }
                }
                else
                {
                    // it is name of executable on the command line
                    StringBuilder sb = new StringBuilder("run");
                    for(i=0;i<commandLineArguments.Length;i++)
                    {
                        sb.Append(' ');
                        string arg = commandLineArguments[i];
                        if( arg.IndexOf(' ') != -1 )
                        {
                            // argument contains spaces, need to quote it
                            sb.Append('\"').Append(arg).Append('\"');
                        }
                        else
                        {
                            sb.Append(arg);
                        }
                    }
                    startupCommands.Add(sb.ToString());
                }

                initialCommands = (string[])startupCommands.ToArray(typeof(string));
            }

            
            this.IO = new MDbgIO(this,initialCommands);

            MdbgCommands.Shell = this;

            m_debugger = new MDbgEngine();
            MdbgCommands.Initialize();

            OnCommandExecuted+=new CommandExecutedEventHandler(MdbgCommands.WhenHandler);

            LoadPlatformDependentExtension();
            
            ProcessAutoExec();
        }

        protected virtual void PrintStartupLogo()
        {
            string myVersion = GetBinaryVersion();
            
            IO.WriteOutput(MDbgOutputConstants.StdOutput,
                           "MDbg (Managed debugger) v"+myVersion+" started.");
            IO.WriteOutput(MDbgOutputConstants.StdOutput,
                           "Copyright (C) Microsoft Corporation. All rights reserved.");
            IO.WriteOutput(MDbgOutputConstants.StdOutput,
                           "\nFor information about commands type \"help\";\nto exit program type \"quit\".\n");

            // Check for and output any debugging warnings
            try {
                Debugger.Processes.DefaultLocalDebugger.CanLaunchOrAttach(0, false );
            } catch( System.Runtime.InteropServices.COMException e ) {
                IO.WriteOutput( MDbgOutputConstants.StdOutput, "WARNING: " + e.Message + "\n" );
            }

        }

        public virtual void DisplayCurrentLocation()
        {
            if (!Debugger.Processes.HaveActive)
            {
                CommandBase.WriteOutput("STOP: Process Exited");
                return; // don't try to display current location
            }
            else
            {
                Debug.Assert(Debugger.Processes.HaveActive);
                Object stopReason = Debugger.Processes.Active.StopReason;
                Type stopReasonType = stopReason.GetType();
                if(stopReasonType == typeof(StepCompleteStopReason))
                {
                    // just ignore those
                }
                else if(stopReasonType == typeof(ThreadCreatedStopReason))
                {
                    CommandBase.WriteOutput("STOP: Thread Created");
                }
                else if(stopReasonType == typeof(BreakpointHitStopReason))
                {
                    MDbgBreakpoint b = (stopReason as BreakpointHitStopReason).Breakpoint;
                    if(b.Number==0)                     // specal case to keep compatibility with our test scripts.
                    {
                        CommandBase.WriteOutput("STOP: Breakpoint Hit");
                    }
                    else
                    {
                        CommandBase.WriteOutput(String.Format(CultureInfo.InvariantCulture, "STOP: Breakpoint {0} Hit", new Object[]{b.Number}));
                    }
                }
                    
                else if(stopReasonType == typeof(ExceptionThrownStopReason))
                {
                    ExceptionThrownStopReason ex = (ExceptionThrownStopReason) stopReason;
                    CommandBase.WriteOutput("STOP: Exception thrown");
                    PrintCurrentException();
                    if(Debugger.Options.StopOnExceptionEnhanced)
                    {
                        // when we are in ExceptionEnhanced mode, we print more information
                        CommandBase.WriteOutput("\tOffset:    " + ex.Offset);
                        CommandBase.WriteOutput("\tEventType: " + ex.EventType);
                        CommandBase.WriteOutput("\tIntercept: " + (ex.Flags!=0));
                    }
                }

                else if(stopReasonType == typeof(UnhandledExceptionThrownStopReason))
                {
                    CommandBase.WriteOutput("STOP: Unhandled Exception thrown");
                    PrintCurrentException();
                    CommandBase.WriteOutput("");
                    CommandBase.WriteOutput("This is unhandled exception, continuing will end the process");
                }
                    
                else if(stopReasonType == typeof(ExceptionUnwindStopReason))
                {
                    CommandBase.WriteOutput("STOP: Exception unwind");
                    CommandBase.WriteOutput("EventType: " + (stopReason as ExceptionUnwindStopReason).EventType);
                }

                else if(stopReasonType == typeof(ModuleLoadedStopReason))
                {
                    CommandBase.WriteOutput("STOP: Module loaded: "+(stopReason as ModuleLoadedStopReason).Module.CorModule.Name);
                }
                else if(stopReasonType == typeof(AssemblyLoadedStopReason))
                {
                    CommandBase.WriteOutput("STOP: Assembly loaded: "+(stopReason as AssemblyLoadedStopReason).Assembly.Name);
                }
                else if(stopReasonType == typeof(MDANotificationStopReason))
                {
                    CorMDA mda = (stopReason as MDANotificationStopReason).CorMDA;

                    CommandBase.WriteOutput("STOP: MDANotification");
                    CommandBase.WriteOutput("Name="+mda.Name);
                    CommandBase.WriteOutput("XML="+mda.XML);
                }
                else if(stopReasonType == typeof(MDbgErrorStopReason))
                {
                    
                    CommandBase.WriteOutput("STOP: MdbgErorr: "+(stopReason as MDbgErrorStopReason).ExceptionThrown.Message);
#if DEBUG
                    CommandBase.WriteOutput((stopReason as MDbgErrorStopReason).ExceptionThrown.ToString());
#endif
                }
                else
                {
                    CommandBase.WriteOutput("STOP " + Debugger.Processes.Active.StopReason);
                }
            }

            if(!Debugger.Processes.Active.Threads.HaveActive)
            {
                return;                                     // we won't try to show current location
            }
            
            MDbgThread thr = Debugger.Processes.Active.Threads.Active;
            
            MDbgSourcePosition pos =thr.CurrentSourcePosition;
            if(pos==null)
            {
                MDbgFrame f = thr.CurrentFrame;
                if( f.IsManaged )
                {
                    CorDebugMappingResult mappingResult;
                    uint ip;
                    f.CorFrame.GetIP(out ip, out mappingResult);
                    string s = "IP: " + ip + " @ " + f.Function.FullName + " - " + mappingResult;
                    CommandBase.WriteOutput(s);
                }
                else
                {
                    CommandBase.WriteOutput("<Located in native code.>");
                }
            }
            else
            {
                string fileLoc = FileLocator.GetFileLocation(pos.Path);
                if(fileLoc==null)
                {
	                // Using the full path makes debugging output inconsistant during automated test runs.
	                // For testing purposes we'll get rid of them.
                    //CommandBase.WriteOutput("located at line "+pos.Line + " in "+ pos.Path);
                    CommandBase.WriteOutput("located at line "+pos.Line + " in "+ System.IO.Path.GetFileName(pos.Path));
                }   
                else
                {
                    IMDbgSourceFile file = SourceFileMgr.GetSourceFile(fileLoc);
                    string prefixStr=pos.Line.ToString(CultureInfo.InvariantCulture)+":";

                    if( pos.Line<1 || pos.Line>file.Count )
                    {
                        CommandBase.WriteOutput("located at line "+pos.Line + " in "+ pos.Path);
                        throw new MDbgShellException(string.Format("Could not display current location; file {0} doesn't have line {1}.",
                                                                   file.Path, pos.Line) );
                    }
                    Debug.Assert( (pos.Line>0) && (pos.Line<=file.Count) );
                    string lineContent = file[pos.Line];
                    
                    if( pos.StartColumn==0 && pos.EndColumn==0
                        || !(CommandBase.Shell.IO is IMDbgIO2) ) // or we don't have support for IMDbgIO2
                    {
                        // we don't know location in the line
                        CommandBase.Shell.IO.WriteOutput(MDbgOutputConstants.StdOutput, prefixStr+lineContent);
                    }
                    else
                    {
                        int hiStart;
                        if( pos.StartColumn>0 )
                        {
                            hiStart = pos.StartColumn-1;
                        }
                        else
                        {
                            hiStart=0;
                        }

                        int hiLen;
                        if( pos.EndColumn==0                   // we don't know ending position
                            || (pos.EndLine > pos.StartLine) ) // multi-line statement, select whole 1st line
                        {
                            hiLen = lineContent.Length;
                        }
                        else
                        {
                            hiLen = pos.EndColumn-1 - hiStart;
                        }
                        Debug.Assert(CommandBase.Shell.IO is IMDbgIO2); // see if condition above
                        (CommandBase.Shell.IO as IMDbgIO2).WriteOutput(MDbgOutputConstants.StdOutput, prefixStr+lineContent,
                                                         hiStart+prefixStr.Length,hiLen);
                    }
                }
            }
        }


        public virtual void ReportException(Exception e)
        {
            if( e == null )
            {
                // If we don't know anything about the exception, report a standard error
                // This should only be used for handling Non-CLS compliant exceptions.  The
                // FxCop rule CatchNonClsCompliantExceptionsInGeneralHandlers requires that
                // we catch non-CLS exceptions whenever we are catching all other exceptions.
                // Ideally, we should just be more selective in the errors we throw and catch
                IO.WriteOutput(MDbgOutputConstants.StdError,"An Unknown error has occurred." );
            }
            else if( e is ThreadInterruptedException )
            {
                IO.WriteOutput(MDbgOutputConstants.StdError,"Interrupted!");
            }
            else
            {
                IO.WriteOutput(MDbgOutputConstants.StdError,e.GetBaseException().Message);
                if(MdbgCommands.AssertOnErrors)
                {
                    IO.WriteOutput(MDbgOutputConstants.StdError,e.GetBaseException().ToString());
                    Debug.Assert(false,"Failure. Assert on Error turned on (mode ae on)\n"+e.GetBaseException().ToString());
                }
            }
        }

        //////////////////////////////////////////////////////////////////////////////////
        //
        // Helper methods
        //
        //////////////////////////////////////////////////////////////////////////////////

        protected static string GetBinaryVersion()
        {
            string assemblyName = Assembly.GetAssembly(typeof(MDbgShell)).Location;
            FileVersionInfo fvi = FileVersionInfo.GetVersionInfo(assemblyName);
            string myVersion=fvi.FileVersion;
            return myVersion;
        }

        // loads extension MdbgDis located next to the debugger.
        protected static bool LoadPlatformDependentExtension()
        {
            const string NativeDasmExtName = "Mdbgdis.dll";
            
            string extdir = AppDomain.CurrentDomain.BaseDirectory;

            Assembly asext = null;
            try
            {
                asext = Assembly.LoadFrom(extdir+Path.DirectorySeparatorChar+NativeDasmExtName);
            }
            catch(FileNotFoundException)
            {
                // we could not find our disassembler.
                return false;
            }
            catch(FileLoadException)
            {
                // we could not load the disassembler.
                return false;
            }
                
            Debug.Assert(asext!=null);


            const string ExtMethod = "LoadExtensionSilently";
            const string ExtNameSpace = "Microsoft.Samples.Tools.Mdbg.Extension.";

            string assemblyName = asext.GetName().Name;
            // get assemblyName into the correct format (see "load" command for more details")
            assemblyName = assemblyName.Substring(0,1).ToUpper(CultureInfo.CurrentUICulture) 
                + assemblyName.Substring(1,assemblyName.Length-1).ToLower(CultureInfo.CurrentUICulture);

            string extType = ExtNameSpace + assemblyName + "Extension";

            Type ext = asext.GetType(extType);

            if(ext==null)
                return false; // we could not load it

            MethodInfo mi = ext.GetMethod(ExtMethod);
            if(mi==null)
                return false; // we could not load it
            mi.Invoke(null,null);
            return true;
        }

        //////////////////////////////////////////////////////////////////////////////////
        //
        // Private implementation
        //
        //////////////////////////////////////////////////////////////////////////////////

        private int RunInputLoop()
        {
            // Run the event loop
            string input;
            IMDbgCommand cmd=null;
            string cmdArgs=null;
            
            int stopCount = -1;
            while(m_run && IO.ReadCommand(out input))
            {
                try 
                {
                    if(Debugger.Processes.HaveActive)
                    {
                        stopCount = Debugger.Processes.Active.StopCounter;
                    }
                    
                    if(input.Length == 0)
                    {
                        if(cmd==null || !cmd.IsRepeatable)
                        {
                            continue;
                        }
                    }
                    else
                    {
                        Commands.ParseCommand(input,out cmd,out cmdArgs);
                    }
                    cmd.Execute(cmdArgs);

                    int newStopCount = Debugger.Processes.HaveActive?Debugger.Processes.Active.StopCounter:Int32.MaxValue;
                    bool movementCommand=newStopCount>stopCount;
                    stopCount = newStopCount;

                    if(OnCommandExecuted!=null)
                    {
                        OnCommandExecuted(this,new CommandExecutedEventArgs(this,cmd,cmdArgs,movementCommand));
                    }

                    newStopCount = Debugger.Processes.HaveActive?Debugger.Processes.Active.StopCounter:Int32.MaxValue;
                    movementCommand = newStopCount>stopCount;

                    while(newStopCount>stopCount)
                    {
                        stopCount=newStopCount;
                        
                        if(OnCommandExecuted!=null)
                        {
                            OnCommandExecuted(this,new CommandExecutedEventArgs(this,null,null,movementCommand));
                        }

                        newStopCount = Debugger.Processes.HaveActive?Debugger.Processes.Active.StopCounter:Int32.MaxValue;
                        movementCommand = newStopCount>stopCount;
                    }
                    stopCount = newStopCount;
                }
                catch(Exception e)
                {
                    ReportException(e);
                }
            } // end while
            return m_exitCode;
        }

		// @TODO - This code is really fragile and specific to our testing setup.
		// At some point maybe we should reduce the output to something smaller and more user-friendly
        private void PrintCurrentException()
        {
            MDbgValue ex = Debugger.Processes.Active.Threads.Active.CurrentException;
            CommandBase.WriteOutput("Exception="+ex.GetStringValue(0));
            foreach(MDbgValue f in ex.GetFields())
            {
                string outputType;
                string outputValue;
                
                if(f.Name == "_xptrs" || f.Name == "_xcode" || f.Name == "_stackTrace" ||
                    f.Name == "_remoteStackTraceString" || f.Name == "_remoteStackIndex" ||
                    f.Name == "_exceptionMethodString")
                {
                    outputType = MDbgOutputConstants.Ignore;
                }
                else
                {
                    outputType = MDbgOutputConstants.StdOutput;
                }

                outputValue = f.GetStringValue(0);
                // remove new line characters in string
                if (outputValue != null && (f.Name == "_exceptionMethodString" || f.Name == "_remoteStackTraceString"))
                {
                    outputValue = outputValue.Replace('\n', '#');
                }
                
                CommandBase.WriteOutput(outputType,"\t"+f.Name +"="+outputValue);
            }
        }


        private void ProcessAutoExec()
        {
            const string autoLoadListFileName = "mdbgAutoExec.txt";
            string mdbgDirectory = AppDomain.CurrentDomain.BaseDirectory;

            StreamReader sr = null;
            try {
                sr = new StreamReader(mdbgDirectory + Path.DirectorySeparatorChar + autoLoadListFileName);

                string s = sr.ReadLine();
                while(s!=null)
                {
                    try {
                        if( (s.Length!=0)
                            && !s.StartsWith("#") )
                            CommandBase.ExecuteCommand(s);
                    }
                    catch(Exception e)
                    {
                        ReportException(e);
                    }
                    s = sr.ReadLine();
                }
            }
            catch(System.IO.FileNotFoundException )
            {
                // we'll ignore this error, since it means that we just don't have the autoExec file.
            }
            catch(System.IO.IOException ex)
            {
                // we'll ignore IO exceptions
                ReportException(ex);
            }
            finally
            {
                if( sr!=null )
                    sr.Close(); // free resources in advance
            }

        }

        //////////////////////////////////////////////////////////////////////////////////
        //
        // Variables
        //
        //////////////////////////////////////////////////////////////////////////////////

        private bool m_run = true;
        private bool m_nologo = false;
        private int m_verbosity = 2;
        private int  m_exitCode = 0; // exit code when we quit.
        
        private MDbgEngine m_debugger = null;
        private IMDbgFileLocator m_fileLocator = new MDbgFileLocator();

        private IMDbgIO m_io;
        private IMDbgCommandCollection m_commands = new MDbgCommandSetCollection();
        private MDbgSourceFileMgr m_sourceFileMgr = new MDbgSourceFileMgr();
        private Hashtable m_properties = new Hashtable();
    }

    
    //////////////////////////////////////////////////////////////////////////////////
    //
    // MDbgOutputImpl
    //
    //////////////////////////////////////////////////////////////////////////////////

    public class MDbgIO : IMDbgIO, IMDbgIO2
    {

        public MDbgIO(MDbgShell shell,string[] startupCommands)
        {
            Debug.Assert(shell!=null);
            m_shell = shell;

            Console.CancelKeyPress += new ConsoleCancelEventHandler(this.ConsoleBreakHandler);

            if( startupCommands!=null && startupCommands.Length>0)
            {
                m_startupCommands = new Queue( startupCommands );
            }
        }

        public virtual void WriteOutput(string outputType,string text)
        {
            WriteOutput(outputType,text,0,0);
        }

        public virtual void WriteOutput(string outputType,string text, int hilightStart, int hilightLen)
        {
            Debug.Assert(null!=outputType && null!=text);

            if( outputType.Equals(MDbgOutputConstants.StdError) )
            {
                if (m_shell.Verbosity >= 0)
                {
                    string s = TEXT_ERROR + text;
                    Display(s,HighlighType.Error,0,s.Length);
                }
            }
            else if (m_shell.Verbosity >= 2)
            {
                Debug.Assert(hilightStart>=0);
                Debug.Assert(hilightLen>=0);
                if(hilightStart<0)
                {
                    hilightStart=0;
                }
                if(hilightLen<0)
                {
                    hilightLen=0;
                }
                Display(text,HighlighType.StatementLocation,hilightStart,hilightLen);
            }
        }

        public virtual bool ReadCommand(out string cmd)
        {
            if(m_startupCommands!=null)
            {
                string c = (string) m_startupCommands.Dequeue();
                if(m_startupCommands.Count==0)
                {
                    m_startupCommands=null;
                }
                cmd = c;
                WriteOutput(MDbgOutputConstants.StdOutput,c);
                return true;
            }
            
            retry:
            m_isConsoleBreakHandlerExecuted = false;
            WriteMdbgPrompt(true);
            cmd = Console.ReadLine();
            if( cmd==null )
            {
                Thread.Sleep(100);

                if( m_isConsoleBreakHandlerExecuted ) 
                    // means we have not hit an EOF.
                    goto retry;
            }
            
            return (cmd==null?false:true);
        }
        
        //////////////////////////////////////////////////////////////////////////////////
        //
        // Private Implementation part
        //
        //////////////////////////////////////////////////////////////////////////////////

        private enum HighlighType 
        {
            StatementLocation,
            Error,
            None
        }

        private void Display(string text,HighlighType ht, int highlightStart,int highlighLen)
        {
            if( highlighLen==0 || highlightStart>=text.Length)
            {
                ConsoleColor fc = Console.ForegroundColor;
                Console.ForegroundColor = ConsoleColor.White;
                Console.WriteLine(text);
                Console.ForegroundColor = fc;
            }
            else
            {
                Console.Write(text.Substring(0,highlightStart));
                ConsoleColor fc = Console.ForegroundColor;
                ConsoleColor bc = Console.BackgroundColor;
                
                switch(ht)
                {
                case HighlighType.StatementLocation:
                    Console.ForegroundColor=ConsoleColor.Yellow;
                    break;
                case HighlighType.Error:
                    Console.ForegroundColor=ConsoleColor.Red;
                    break;
                default:
                    Debug.Assert(false);
                    throw new InvalidOperationException();
                }
                int l=highlightStart+highlighLen;
                if( l>text.Length )
                {
                    highlighLen=text.Length-highlightStart;
                }
                Console.Write(text.Substring(highlightStart,highlighLen));
                Console.ForegroundColor=fc;
                Console.BackgroundColor=bc;
                if( highlightStart+highlighLen<text.Length )
                {
                    Console.Write(text.Substring(highlightStart+highlighLen));
                }
                Console.WriteLine();
            }
        }

        protected bool m_isConsoleBreakHandlerExecuted = false;
        
        private void ConsoleBreakHandler(Object sender,ConsoleCancelEventArgs e)
        {
            // BUG! When Control+C is pressed, the Console.Readline() returns immediatelly with
            // null. There is no way how to distinguish between EOF and Ctrl+C. Therefore we set
            // a m_isConsoleBreakHandlerExecuted to true whenever the handler is executed.
            // The code around Console.ReadLine() check when null is returned if the
            // m_isConsoleBreakHandlerExecuted flag is set. If yes, we know that this is caused by
            // Ctrl+C, the code clrears the flag and repeats the read.
            
            m_isConsoleBreakHandlerExecuted = true;
            
            switch(e.SpecialKey)
            {
            case ConsoleSpecialKey.ControlBreak:
                Console.WriteLine();
                WriteOutput(MDbgOutputConstants.StdError,"Immediate debugger termination reqested through <Ctrl+Break>");
                WriteOutput(MDbgOutputConstants.StdError,"To break into debugger use Ctrl+C instead.");
                //
                // When ControlBreak is pressed, we cannot set e.Cancel=true....
                // 
                break;
            case ConsoleSpecialKey.ControlC:
                try
                {
                    WriteOutput(MDbgOutputConstants.StdOutput,"\n<Ctrl+C>");
                    MDbgProcess p = m_shell.Debugger.Processes.Active;
                    if( p.IsRunning )
                    {
                        p.AsyncStop().WaitOne();
                    }
                }
                catch
                {
                }
                e.Cancel = true;
                break;
            default:
                break;
            }
        }

        private void WriteMdbgPrompt(bool processStopped)
        {
            string s;
            if(processStopped)
            {
                MDbgProcess p=null;
                if(m_shell.Debugger.Processes.HaveActive)
                {
                    p = m_shell.Debugger.Processes.Active;
                    if(p.Threads.HaveActive)
                    {
                        s = "[p#:"+p.Number+", t#:"+p.Threads.Active.Number+"] ";
                    }
                    else
                    {
                        s = "[p#:"+p.Number+", t#:no active thread] ";
                    }
                }
                else
                {
                    s = "";
                }
            }
            else
            {
                s = "[process running] ";
            }
            Console.Write(s + TEXT_PROMPT_STRING);
        }


        private const string TEXT_PROMPT_STRING="mdbg> ";
        private const string TEXT_ERROR="Error: ";
        private MDbgShell m_shell;
        protected Queue m_startupCommands = null;
    }

    //////////////////////////////////////////////////////////////////////////////////
    //
    // MDbgCommandSetImpl
    //
    //////////////////////////////////////////////////////////////////////////////////

    public class MDbgCommandSetCollection : IMDbgCommandCollection
    {
        public MDbgCommandSetCollection()
        {
        }

        public void Add(IMDbgCommand command)
        {
            
            if (command ==null)
            {
                Debug.Assert(false);
                throw new Exception();
            }            

            if (command.MinimumAbbrev > command.CommandName.Length)
            {
                throw new MDbgShellException("Cannot add command '"+command.CommandName+"'. Abbreviation is " + 
                    command.MinimumAbbrev +" characters. Can't be more than "+ command.CommandName.Length);
            }
            

            // extensions are allowed to override a command, i.e. this means we will delete previously defined command
            // with such name.
            foreach(IMDbgCommand c in m_commands)
            {
                if(c.CommandName.Equals(command.CommandName))
                {
                    m_commands.Remove(c);
                    break;                                  // there should never be 2 command with same name in a collection
                                                            // therefore we can break-out from the search.
                }
            }

            m_commands.Add(command);
            m_needSorting = true;
        }

        public IMDbgCommand Lookup(string commandName)
        {
            ArrayList al = new ArrayList();
            foreach(IMDbgCommand c in m_commands)
            {
                if(commandName.Length<c.MinimumAbbrev)
                {
                    continue;
                }
                if(String.Compare(c.CommandName,0,commandName,0,commandName.Length,true,
                                  CultureInfo.CurrentUICulture // command names could get localized in the future
                                  )==0)
                {
                    al.Add(c);
                }
            }
            if(al.Count==0)
            {
                throw new MDbgShellException("Command '"+commandName+"' not found.");
            }
            else if(al.Count==1)
            {
                return (IMDbgCommand)al[0];
            }
            else
            {
                StringBuilder s=new StringBuilder("Command prefix too short. \nPossible completitions:");
                foreach(IMDbgCommand c in al) 
                {
                    s.Append("\n").Append(c.CommandName);
                }
                throw new MDbgShellException(s.ToString()); 
            }
        }

        public IEnumerator GetEnumerator()
        {
            if(m_needSorting)
            {
                m_needSorting=false;
                m_commands.Sort();
            }

            return m_commands.GetEnumerator();
        }

        public void ParseCommand(string commandLineText, out IMDbgCommand command,out string commandArguments)
        {
            commandLineText = commandLineText.Trim();
            int n = commandLineText.Length;
            int i = 0;
            while(i<n && ! Char.IsWhiteSpace(commandLineText,i))
            {
                i++;
            }
            string cmdName = commandLineText.Substring(0,i);
            commandArguments = commandLineText.Substring(i).Trim();
            command = Lookup(cmdName);
        }

        private ArrayList m_commands = new ArrayList();
        private bool m_needSorting = false;
    }


    //////////////////////////////////////////////////////////////////////////////////
    //
    // MDbgFileLocatorImpl
    //
    //////////////////////////////////////////////////////////////////////////////////

    public class MDbgFileLocator : IMDbgFileLocator
    {
        internal MDbgFileLocator()
        {
        }

        public string Path 
        {
            get
            {
                return m_srcPath;
            }
            set
            {
                foreach(string pathPart in GetPathComponents(value))
                {
                    if(!Directory.Exists(pathPart))
                    {
                        throw new MDbgShellException("path doesn't exist: '"+pathPart+"'");
                    }
                }
                m_srcPath = value;
                m_fileLocations.Clear();
            }
        }

        public string GetFileLocation(string file)
        {
            string idx = String.Intern(file);
            if(m_fileLocations.Contains(idx))
            {
                return (string) m_fileLocations[idx];
            }

            string realPath=null;
            
            if(File.Exists(file))
            {
                realPath = file;
            }
            else
            {
                foreach(string pathPart in GetPathComponents(m_srcPath))
                {
                    string filePath = file;
                    do
                    {
                        realPath = pathPart+"\\"+filePath;
                        if(File.Exists(realPath))
                            goto Found;

                        int i = filePath.IndexOfAny(new char[]{
                            System.IO.Path.DirectorySeparatorChar,System.IO.Path.AltDirectorySeparatorChar});
                        if(i!=-1)
                            filePath = filePath.Substring(i+1);
                        else
                            break;
                    }
                    while(true);
                }
                realPath = null;
            }
            Found:
            m_fileLocations.Add(idx,realPath);
            return realPath;
        }

        public void Associate(string originalName, string newName)
        {
            Debug.Assert( originalName!=null && originalName.Length>0);
            Debug.Assert( newName!=null && newName.Length>0 );
            
            string idx = String.Intern(originalName);
			if(m_fileLocations.Contains(idx))
				m_fileLocations.Remove(idx);
            m_fileLocations.Add(idx, newName);
        }

        protected string[] GetPathComponents(string path)
        {
            if(path==null)
            {
                return new string[0];
            }
            string[] strs = path.Split(new char[]{';'});
            return strs;
        }

        private Hashtable m_fileLocations = new Hashtable();
        private string m_srcPath;
    }
}
